import unicorn

from ....platforms import Architecture, Byteorder
from .machdef import UnicornMachineDef


class AMD64MachineDef(UnicornMachineDef):
    """Unicorn machine definition for amd64"""

    byteorder = Byteorder.LITTLE

    uc_arch = unicorn.UC_ARCH_X86
    uc_mode = unicorn.UC_MODE_64

    def __init__(self):
        self._registers = {
            # *** General Purpose Registers ***
            "rax": unicorn.x86_const.UC_X86_REG_RAX,
            "eax": unicorn.x86_const.UC_X86_REG_EAX,
            "ax": unicorn.x86_const.UC_X86_REG_AX,
            "al": unicorn.x86_const.UC_X86_REG_AL,
            "ah": unicorn.x86_const.UC_X86_REG_AH,
            "rbx": unicorn.x86_const.UC_X86_REG_RBX,
            "ebx": unicorn.x86_const.UC_X86_REG_EBX,
            "bx": unicorn.x86_const.UC_X86_REG_BX,
            "bl": unicorn.x86_const.UC_X86_REG_BL,
            "bh": unicorn.x86_const.UC_X86_REG_BH,
            "rcx": unicorn.x86_const.UC_X86_REG_RCX,
            "ecx": unicorn.x86_const.UC_X86_REG_ECX,
            "cx": unicorn.x86_const.UC_X86_REG_CX,
            "cl": unicorn.x86_const.UC_X86_REG_CL,
            "ch": unicorn.x86_const.UC_X86_REG_CH,
            "rdx": unicorn.x86_const.UC_X86_REG_RDX,
            "edx": unicorn.x86_const.UC_X86_REG_EDX,
            "dx": unicorn.x86_const.UC_X86_REG_DX,
            "dl": unicorn.x86_const.UC_X86_REG_DL,
            "dh": unicorn.x86_const.UC_X86_REG_DH,
            "r8": unicorn.x86_const.UC_X86_REG_R8,
            "r8d": unicorn.x86_const.UC_X86_REG_R8D,
            "r8w": unicorn.x86_const.UC_X86_REG_R8W,
            "r8b": unicorn.x86_const.UC_X86_REG_R8B,
            "r9": unicorn.x86_const.UC_X86_REG_R9,
            "r9d": unicorn.x86_const.UC_X86_REG_R9D,
            "r9w": unicorn.x86_const.UC_X86_REG_R9W,
            "r9b": unicorn.x86_const.UC_X86_REG_R9B,
            "r10": unicorn.x86_const.UC_X86_REG_R10,
            "r10d": unicorn.x86_const.UC_X86_REG_R10D,
            "r10w": unicorn.x86_const.UC_X86_REG_R10W,
            "r10b": unicorn.x86_const.UC_X86_REG_R10B,
            "r11": unicorn.x86_const.UC_X86_REG_R11,
            "r11d": unicorn.x86_const.UC_X86_REG_R11D,
            "r11w": unicorn.x86_const.UC_X86_REG_R11W,
            "r11b": unicorn.x86_const.UC_X86_REG_R11B,
            "r12": unicorn.x86_const.UC_X86_REG_R12,
            "r12d": unicorn.x86_const.UC_X86_REG_R12D,
            "r12w": unicorn.x86_const.UC_X86_REG_R12W,
            "r12b": unicorn.x86_const.UC_X86_REG_R12B,
            "r13": unicorn.x86_const.UC_X86_REG_R13,
            "r13d": unicorn.x86_const.UC_X86_REG_R13D,
            "r13w": unicorn.x86_const.UC_X86_REG_R13W,
            "r13b": unicorn.x86_const.UC_X86_REG_R13B,
            "r14": unicorn.x86_const.UC_X86_REG_R14,
            "r14d": unicorn.x86_const.UC_X86_REG_R14D,
            "r14w": unicorn.x86_const.UC_X86_REG_R14W,
            "r14b": unicorn.x86_const.UC_X86_REG_R14B,
            "r15": unicorn.x86_const.UC_X86_REG_R15,
            "r15d": unicorn.x86_const.UC_X86_REG_R15D,
            "r15w": unicorn.x86_const.UC_X86_REG_R15W,
            "r15b": unicorn.x86_const.UC_X86_REG_R15B,
            "rdi": unicorn.x86_const.UC_X86_REG_RDI,
            "edi": unicorn.x86_const.UC_X86_REG_EDI,
            "di": unicorn.x86_const.UC_X86_REG_DI,
            "dil": unicorn.x86_const.UC_X86_REG_DIL,
            "rsi": unicorn.x86_const.UC_X86_REG_RSI,
            "esi": unicorn.x86_const.UC_X86_REG_ESI,
            "si": unicorn.x86_const.UC_X86_REG_SI,
            "sil": unicorn.x86_const.UC_X86_REG_SIL,
            "rsp": unicorn.x86_const.UC_X86_REG_RSP,
            "esp": unicorn.x86_const.UC_X86_REG_ESP,
            "sp": unicorn.x86_const.UC_X86_REG_SP,
            "spl": unicorn.x86_const.UC_X86_REG_SPL,
            "rbp": unicorn.x86_const.UC_X86_REG_RBP,
            "ebp": unicorn.x86_const.UC_X86_REG_EBP,
            "bp": unicorn.x86_const.UC_X86_REG_BP,
            "bpl": unicorn.x86_const.UC_X86_REG_BPL,
            # *** Instruction Pointer ***
            "rip": unicorn.x86_const.UC_X86_REG_RIP,
            "eip": unicorn.x86_const.UC_X86_REG_EIP,
            "ip": unicorn.x86_const.UC_X86_REG_IP,
            # *** Flags register ***
            "rflags": unicorn.x86_const.UC_X86_REG_RFLAGS,
            "eflags": unicorn.x86_const.UC_X86_REG_EFLAGS,
            "flags": unicorn.x86_const.UC_X86_REG_FLAGS,
            # *** Segment Registers ***
            "cs": unicorn.x86_const.UC_X86_REG_CS,
            "ds": unicorn.x86_const.UC_X86_REG_DS,
            "es": unicorn.x86_const.UC_X86_REG_ES,
            "fs": unicorn.x86_const.UC_X86_REG_FS,
            "gs": unicorn.x86_const.UC_X86_REG_GS,
            "ss": unicorn.x86_const.UC_X86_REG_SS,
            # *** Control Registers ***
            "cr0": unicorn.x86_const.UC_X86_REG_CR0,
            "cr1": unicorn.x86_const.UC_X86_REG_CR1,
            "cr2": unicorn.x86_const.UC_X86_REG_CR2,
            "cr3": unicorn.x86_const.UC_X86_REG_CR3,
            "cr4": unicorn.x86_const.UC_X86_REG_CR4,
            "cr8": unicorn.x86_const.UC_X86_REG_INVALID,
            # *** Debug Registers ***
            "dr0": unicorn.x86_const.UC_X86_REG_DR0,
            "dr1": unicorn.x86_const.UC_X86_REG_DR1,
            "dr2": unicorn.x86_const.UC_X86_REG_DR2,
            "dr3": unicorn.x86_const.UC_X86_REG_DR3,
            "dr6": unicorn.x86_const.UC_X86_REG_DR6,
            "dr7": unicorn.x86_const.UC_X86_REG_DR7,
            "dr8": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr9": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr10": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr11": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr12": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr13": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr14": unicorn.x86_const.UC_X86_REG_INVALID,
            "dr15": unicorn.x86_const.UC_X86_REG_INVALID,
            # *** Descriptor Table Registers ***
            "gdtr": unicorn.x86_const.UC_X86_REG_GDTR,
            "idtr": unicorn.x86_const.UC_X86_REG_IDTR,
            "ldtr": unicorn.x86_const.UC_X86_REG_LDTR,
            # *** Task Register ***
            "tr": unicorn.x86_const.UC_X86_REG_TR,
            # *** x87 registers ***
            # NOTE: x87 is supported by Unicorn, but not by SmallWorld.
            # Values are represented as tuples (exponent: int, mantissa: int).
            # If you need x87 support, open a ticket.
            "fpr0": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr1": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr2": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr3": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr4": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr5": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr6": unicorn.x86_const.UC_X86_REG_INVALID,
            "fpr7": unicorn.x86_const.UC_X86_REG_INVALID,
            # x87 Control Register
            "fctrl": unicorn.x86_const.UC_X86_REG_FPCW,
            # x87 Status Register
            "fstat": unicorn.x86_const.UC_X86_REG_FPCW,
            # x87 Tag Register
            "ftag": unicorn.x86_const.UC_X86_REG_FPTAG,
            # x87 Last Instruction Register
            "fip": unicorn.x86_const.UC_X86_REG_FIP,
            # x87 Last Operand Pointer
            "fdp": unicorn.x86_const.UC_X86_REG_FDP,
            # x87 Last Opcode
            "fop": unicorn.x86_const.UC_X86_REG_FOP,
            # NOTE: Docs disagree on the format of fip and fdp.
            # One source describes them as 48-bit offset-plus-segment,
            # the other describes them as 64-bit.
            # There may also be separate segment registers.
            # If you care about the x87 debug info, please feel free to update.
            # *** MMX Registers ***
            # NOTE: The MMX registers are aliases for the low 8 bytes of the x87 registers.
            # The two subsystems cannot be used simultaneously.
            "mm0": unicorn.x86_const.UC_X86_REG_MM0,
            "mm1": unicorn.x86_const.UC_X86_REG_MM1,
            "mm2": unicorn.x86_const.UC_X86_REG_MM2,
            "mm3": unicorn.x86_const.UC_X86_REG_MM3,
            "mm4": unicorn.x86_const.UC_X86_REG_MM4,
            "mm5": unicorn.x86_const.UC_X86_REG_MM5,
            "mm6": unicorn.x86_const.UC_X86_REG_MM6,
            "mm7": unicorn.x86_const.UC_X86_REG_MM7,
        }


class AMD64AVX2MachineDef(AMD64MachineDef):
    arch = Architecture.X86_64

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** SSE/AVX registers ***
                "ymm0": unicorn.x86_const.UC_X86_REG_YMM0,
                "xmm0": unicorn.x86_const.UC_X86_REG_XMM0,
                "ymm1": unicorn.x86_const.UC_X86_REG_YMM1,
                "xmm1": unicorn.x86_const.UC_X86_REG_XMM1,
                "ymm2": unicorn.x86_const.UC_X86_REG_YMM2,
                "xmm2": unicorn.x86_const.UC_X86_REG_XMM2,
                "ymm3": unicorn.x86_const.UC_X86_REG_YMM3,
                "xmm3": unicorn.x86_const.UC_X86_REG_XMM3,
                "ymm4": unicorn.x86_const.UC_X86_REG_YMM4,
                "xmm4": unicorn.x86_const.UC_X86_REG_XMM4,
                "ymm5": unicorn.x86_const.UC_X86_REG_YMM5,
                "xmm5": unicorn.x86_const.UC_X86_REG_XMM5,
                "ymm6": unicorn.x86_const.UC_X86_REG_YMM6,
                "xmm6": unicorn.x86_const.UC_X86_REG_XMM6,
                "ymm7": unicorn.x86_const.UC_X86_REG_YMM7,
                "xmm7": unicorn.x86_const.UC_X86_REG_XMM7,
                "ymm8": unicorn.x86_const.UC_X86_REG_YMM8,
                "xmm8": unicorn.x86_const.UC_X86_REG_XMM8,
                "ymm9": unicorn.x86_const.UC_X86_REG_YMM9,
                "xmm9": unicorn.x86_const.UC_X86_REG_XMM9,
                "ymm10": unicorn.x86_const.UC_X86_REG_YMM10,
                "xmm10": unicorn.x86_const.UC_X86_REG_XMM10,
                "ymm11": unicorn.x86_const.UC_X86_REG_YMM11,
                "xmm11": unicorn.x86_const.UC_X86_REG_XMM11,
                "ymm12": unicorn.x86_const.UC_X86_REG_YMM12,
                "xmm12": unicorn.x86_const.UC_X86_REG_XMM12,
                "ymm13": unicorn.x86_const.UC_X86_REG_YMM13,
                "xmm13": unicorn.x86_const.UC_X86_REG_XMM13,
                "ymm14": unicorn.x86_const.UC_X86_REG_YMM14,
                "xmm14": unicorn.x86_const.UC_X86_REG_XMM14,
                "ymm15": unicorn.x86_const.UC_X86_REG_YMM15,
                "xmm15": unicorn.x86_const.UC_X86_REG_XMM15,
            }
        )


class AMD64AVX512MachineDef(AMD64MachineDef):
    arch = Architecture.X86_64_AVX512

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** SSE/AVX registers ***
                "zmm0": unicorn.x86_const.UC_X86_REG_ZMM0,
                "ymm0": unicorn.x86_const.UC_X86_REG_YMM0,
                "xmm0": unicorn.x86_const.UC_X86_REG_XMM0,
                "zmm1": unicorn.x86_const.UC_X86_REG_ZMM1,
                "ymm1": unicorn.x86_const.UC_X86_REG_YMM1,
                "xmm1": unicorn.x86_const.UC_X86_REG_XMM1,
                "zmm2": unicorn.x86_const.UC_X86_REG_ZMM2,
                "ymm2": unicorn.x86_const.UC_X86_REG_YMM2,
                "xmm2": unicorn.x86_const.UC_X86_REG_XMM2,
                "zmm3": unicorn.x86_const.UC_X86_REG_ZMM3,
                "ymm3": unicorn.x86_const.UC_X86_REG_YMM3,
                "xmm3": unicorn.x86_const.UC_X86_REG_XMM3,
                "zmm4": unicorn.x86_const.UC_X86_REG_ZMM4,
                "ymm4": unicorn.x86_const.UC_X86_REG_YMM4,
                "xmm4": unicorn.x86_const.UC_X86_REG_XMM4,
                "zmm5": unicorn.x86_const.UC_X86_REG_ZMM5,
                "ymm5": unicorn.x86_const.UC_X86_REG_YMM5,
                "xmm5": unicorn.x86_const.UC_X86_REG_XMM5,
                "zmm6": unicorn.x86_const.UC_X86_REG_ZMM6,
                "ymm6": unicorn.x86_const.UC_X86_REG_YMM6,
                "xmm6": unicorn.x86_const.UC_X86_REG_XMM6,
                "zmm7": unicorn.x86_const.UC_X86_REG_ZMM7,
                "ymm7": unicorn.x86_const.UC_X86_REG_YMM7,
                "xmm7": unicorn.x86_const.UC_X86_REG_XMM7,
                "zmm8": unicorn.x86_const.UC_X86_REG_ZMM8,
                "ymm8": unicorn.x86_const.UC_X86_REG_YMM8,
                "xmm8": unicorn.x86_const.UC_X86_REG_XMM8,
                "zmm9": unicorn.x86_const.UC_X86_REG_ZMM9,
                "ymm9": unicorn.x86_const.UC_X86_REG_YMM9,
                "xmm9": unicorn.x86_const.UC_X86_REG_XMM9,
                "zmm10": unicorn.x86_const.UC_X86_REG_ZMM10,
                "ymm10": unicorn.x86_const.UC_X86_REG_YMM10,
                "xmm10": unicorn.x86_const.UC_X86_REG_XMM10,
                "zmm11": unicorn.x86_const.UC_X86_REG_ZMM11,
                "ymm11": unicorn.x86_const.UC_X86_REG_YMM11,
                "xmm11": unicorn.x86_const.UC_X86_REG_XMM11,
                "zmm12": unicorn.x86_const.UC_X86_REG_ZMM12,
                "ymm12": unicorn.x86_const.UC_X86_REG_YMM12,
                "xmm12": unicorn.x86_const.UC_X86_REG_XMM12,
                "zmm13": unicorn.x86_const.UC_X86_REG_ZMM13,
                "ymm13": unicorn.x86_const.UC_X86_REG_YMM13,
                "xmm13": unicorn.x86_const.UC_X86_REG_XMM13,
                "zmm14": unicorn.x86_const.UC_X86_REG_ZMM14,
                "ymm14": unicorn.x86_const.UC_X86_REG_YMM14,
                "xmm14": unicorn.x86_const.UC_X86_REG_XMM14,
                "zmm15": unicorn.x86_const.UC_X86_REG_ZMM15,
                "ymm15": unicorn.x86_const.UC_X86_REG_YMM15,
                "xmm15": unicorn.x86_const.UC_X86_REG_XMM15,
                "zmm16": unicorn.x86_const.UC_X86_REG_ZMM16,
                "ymm16": unicorn.x86_const.UC_X86_REG_YMM16,
                "xmm16": unicorn.x86_const.UC_X86_REG_XMM16,
                "zmm17": unicorn.x86_const.UC_X86_REG_ZMM17,
                "ymm17": unicorn.x86_const.UC_X86_REG_YMM17,
                "xmm17": unicorn.x86_const.UC_X86_REG_XMM17,
                "zmm18": unicorn.x86_const.UC_X86_REG_ZMM18,
                "ymm18": unicorn.x86_const.UC_X86_REG_YMM18,
                "xmm18": unicorn.x86_const.UC_X86_REG_XMM18,
                "zmm19": unicorn.x86_const.UC_X86_REG_ZMM19,
                "ymm19": unicorn.x86_const.UC_X86_REG_YMM19,
                "xmm19": unicorn.x86_const.UC_X86_REG_XMM19,
                "zmm20": unicorn.x86_const.UC_X86_REG_ZMM20,
                "ymm20": unicorn.x86_const.UC_X86_REG_YMM20,
                "xmm20": unicorn.x86_const.UC_X86_REG_XMM20,
                "zmm21": unicorn.x86_const.UC_X86_REG_ZMM21,
                "ymm21": unicorn.x86_const.UC_X86_REG_YMM21,
                "xmm21": unicorn.x86_const.UC_X86_REG_XMM21,
                "zmm22": unicorn.x86_const.UC_X86_REG_ZMM22,
                "ymm22": unicorn.x86_const.UC_X86_REG_YMM22,
                "xmm22": unicorn.x86_const.UC_X86_REG_XMM22,
                "zmm23": unicorn.x86_const.UC_X86_REG_ZMM23,
                "ymm23": unicorn.x86_const.UC_X86_REG_YMM23,
                "xmm23": unicorn.x86_const.UC_X86_REG_XMM23,
                "zmm24": unicorn.x86_const.UC_X86_REG_ZMM24,
                "ymm24": unicorn.x86_const.UC_X86_REG_YMM24,
                "xmm24": unicorn.x86_const.UC_X86_REG_XMM24,
                "zmm25": unicorn.x86_const.UC_X86_REG_ZMM25,
                "ymm25": unicorn.x86_const.UC_X86_REG_YMM25,
                "xmm25": unicorn.x86_const.UC_X86_REG_XMM25,
                "zmm26": unicorn.x86_const.UC_X86_REG_ZMM26,
                "ymm26": unicorn.x86_const.UC_X86_REG_YMM26,
                "xmm26": unicorn.x86_const.UC_X86_REG_XMM26,
                "zmm27": unicorn.x86_const.UC_X86_REG_ZMM27,
                "ymm27": unicorn.x86_const.UC_X86_REG_YMM27,
                "xmm27": unicorn.x86_const.UC_X86_REG_XMM27,
                "zmm28": unicorn.x86_const.UC_X86_REG_ZMM28,
                "ymm28": unicorn.x86_const.UC_X86_REG_YMM28,
                "xmm28": unicorn.x86_const.UC_X86_REG_XMM28,
                "zmm29": unicorn.x86_const.UC_X86_REG_ZMM29,
                "ymm29": unicorn.x86_const.UC_X86_REG_YMM29,
                "xmm29": unicorn.x86_const.UC_X86_REG_XMM29,
                "zmm30": unicorn.x86_const.UC_X86_REG_ZMM30,
                "ymm30": unicorn.x86_const.UC_X86_REG_YMM30,
                "xmm30": unicorn.x86_const.UC_X86_REG_XMM30,
                "zmm31": unicorn.x86_const.UC_X86_REG_ZMM31,
                "ymm31": unicorn.x86_const.UC_X86_REG_YMM31,
                "xmm31": unicorn.x86_const.UC_X86_REG_XMM31,
            }
        )



import argparse
import copy
import logging
import typing

logger = logging.getLogger(__name__)


from . import analyses, emulators, exceptions, state

T = typing.TypeVar("T", bound=state.Machine)


def analyze(
    machine: state.Machine,
    asys: typing.List[analyses.Analysis],
) -> None:
    """Run requested analyses on some code

    Arguments:
        machine: A state class from which emulation should begin.
        asys: List of analyses and filters to run
    """
    for a in asys:
        a.run(copy.deepcopy(machine))


def fuzz(
    machine: state.Machine,
    input_callback: typing.Callable,
    crash_callback: typing.Optional[typing.Callable] = None,
    always_validate: bool = False,
    iterations: int = 1,
) -> None:
    """Creates an AFL fuzzing harness.

    Arguments:
        cpu: A state class from which emulation should begin.
        input_callback: This is called for every input. It should map the input
            into the state. It should return true if the input is accepted and
            false if it should be skipped.
        crash_callback: This is called on crashes to validate that we do in
            fact care about this crash.
        always_validate: Call the crash_callback everytime instead of just on
            crashes.
        iterations: How many iterations to run before forking again.
    """

    try:
        import unicornafl
    except ImportError:
        raise RuntimeError(
            "missing `unicornafl` - afl++ must be installed manually from source"
        )

    arg_parser = argparse.ArgumentParser(description="AFL Harness")
    arg_parser.add_argument("input_file", type=str, help="File path AFL will mutate")
    args = arg_parser.parse_args()

    cpu: typing.Optional[state.cpus.CPU] = None
    for c in machine.members(state.cpus.CPU).items():
        if cpu is not None:
            raise exceptions.ConfigurationError("cannot fuzz a multi-core machine")
        cpu = c
    if cpu is None:
        raise exceptions.ConfigurationError("cannot fuzz a zero-core machine")

    emu = emulators.UnicornEmulator(cpu.platform)
    machine.apply(emu)

    exits = []
    code = None
    for code in machine.members(state.memory.code.Executable).items():
        exits.extend([b.stop for b in code.bounds])

    if len(exits) == 0:
        if code is None:
            raise exceptions.ConfigurationError("Cannot fuzz a machine with no code")
        exits.append(code.base + len(code.image))

    unicornafl.uc_afl_fuzz(
        uc=emu.engine,
        input_file=args.input_file,
        place_input_callback=input_callback,
        exits=exits,
        validate_crash_callback=crash_callback,
        always_validate=always_validate,
        persistent_iters=iterations,
    )


__all__ = ["analyze"]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import ArmELSysVModel


class ArmELSysVMemcpy(Memcpy, ArmELSysVModel):
    pass


class ArmELSysVMemmove(Memmove, ArmELSysVModel):
    pass


class ArmELSysVStrcpy(Strcpy, ArmELSysVModel):
    pass


class ArmELSysVStrncpy(Strncpy, ArmELSysVModel):
    pass


class ArmELSysVStrcat(Strcat, ArmELSysVModel):
    pass


class ArmELSysVStrncat(Strncat, ArmELSysVModel):
    pass


class ArmELSysVMemcmp(Memcmp, ArmELSysVModel):
    pass


class ArmELSysVStrncmp(Strncmp, ArmELSysVModel):
    pass


class ArmELSysVStrcmp(Strcmp, ArmELSysVModel):
    pass


class ArmELSysVStrcoll(Strcoll, ArmELSysVModel):
    pass


class ArmELSysVStrxfrm(Strxfrm, ArmELSysVModel):
    pass


class ArmELSysVMemchr(Memchr, ArmELSysVModel):
    pass


class ArmELSysVStrchr(Strchr, ArmELSysVModel):
    pass


class ArmELSysVStrcspn(Strcspn, ArmELSysVModel):
    pass


class ArmELSysVStrpbrk(Strpbrk, ArmELSysVModel):
    pass


class ArmELSysVStrrchr(Strrchr, ArmELSysVModel):
    pass


class ArmELSysVStrspn(Strspn, ArmELSysVModel):
    pass


class ArmELSysVStrstr(Strstr, ArmELSysVModel):
    pass


class ArmELSysVStrtok(Strtok, ArmELSysVModel):
    pass


class ArmELSysVMemset(Memset, ArmELSysVModel):
    pass


class ArmELSysVStrerror(Strerror, ArmELSysVModel):
    pass


class ArmELSysVStrlen(Strlen, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVMemcpy",
    "ArmELSysVMemmove",
    "ArmELSysVStrcpy",
    "ArmELSysVStrncpy",
    "ArmELSysVStrcat",
    "ArmELSysVStrncat",
    "ArmELSysVMemcmp",
    "ArmELSysVStrncmp",
    "ArmELSysVStrcmp",
    "ArmELSysVStrcoll",
    "ArmELSysVStrxfrm",
    "ArmELSysVMemchr",
    "ArmELSysVStrchr",
    "ArmELSysVStrcspn",
    "ArmELSysVStrpbrk",
    "ArmELSysVStrrchr",
    "ArmELSysVStrspn",
    "ArmELSysVStrstr",
    "ArmELSysVStrtok",
    "ArmELSysVMemset",
    "ArmELSysVStrerror",
    "ArmELSysVStrlen",
]



import angr


class SyscallHookProcedure(angr.SimProcedure):
    def run(self):
        # Get the syscall number
        number = self.cc.syscall_num(self.state)

        if number.symbolic:
            raise NotImplementedError(f"Symbolic syscall number {number}")
        number = number.concrete_value

        # See if we have a global handler
        global_func = self.state.scratch.global_syscall_func
        if global_func is not None:
            global_func(self.state, number)

        # See if we have a local handler
        if number in self.state.scratch.syscall_funcs:
            local_func = self.state.scratch.syscall_funcs[number]
            local_func(self.state)

        # Force execution to resume at the syscall exit point.
        self.jump(self.state._ip)


class HookableSimOS(angr.simos.simos.SimOS):
    def syscall(self, state, allow_unsupported=True):
        SYSCALL_CC = angr.calling_conventions.SYSCALL_CC
        arch_name = state.arch.name
        os_name = state.os_name
        if arch_name in SYSCALL_CC:
            if os_name in SYSCALL_CC[arch_name]:
                cc = SYSCALL_CC[arch_name][os_name](state.arch)
            else:
                cc = SYSCALL_CC[arch_name]["default"](state.arch)
        else:
            cc = None

        out = SyscallHookProcedure(
            project=self.project,
            cc=cc,
            prototype=None,
            returns=None,
            is_syscall=True,
            is_stub=False,
            num_args=None,
            display_name=None,
            library_name=None,
            is_function=None,
        )
        out.addr = state._ip
        return out



import abc
import typing

import smallworld.emulators

from .. import state
from . import memory


class Heap(memory.Memory):
    """A heap-like memory region, with convenient operations like allocate and free."""

    @abc.abstractmethod
    def allocate(self, value: state.Value) -> int:
        """Allocate space for and write a value to the heap.

        Arguments:
            value: The value to go on the heap.
        Returns:
            The address at which the value was allocated.
        """
        pass

    @abc.abstractmethod
    def free(self, address: int) -> None:
        """Free space on the heap for previously allocated data.

        Arguements:
            address: The start address of the previously allocated data.
        """
        pass

    def allocate_integer(self, integer: int, size: int, label: str) -> int:
        """Allocate space for and write an integer to the heap.

        Arguments:
            integer: The integer to go on the heap.
            size: The number of bytes to allocate to the integer.
            label: The label to give to the allocated region containing the integer.
        Returns:
            The address at which the integer was allocated.
        """
        value = state.IntegerValue(integer, size, label)
        return self.allocate(value)

    def allocate_bytes(
        self, content: typing.Union[bytes, bytearray], label: typing.Optional[str]
    ) -> int:
        """Allocate space for and write bytes to the heap.

        Arguments:
            content: The bytes to go on the heap.
            label: The label to give to the bytes on the heap.
        Returns:
            The address at which the integer was allocated.
        """
        value = state.BytesValue(content, label)
        return self.allocate(value)

    def allocate_ctype(self, content, label: str) -> int:
        """Allocate space for and write structured bytes to the heap.

        Arguements:
            content: The ctypes-structured data to go on the heap.
            label: The label to give to the data bytes on the heap.

        Returns:
            The address at which the structured data was allocated.
        """
        value = state.Value.from_ctypes(content, label)
        return self.allocate(value)


class BumpAllocator(Heap):
    """A simple bump allocator heap."""

    def allocate(self, value: state.Value) -> int:
        self._is_safe(value)
        offset = self.get_used()
        self[offset] = value
        return self.address + offset

    def free(self, address: int) -> None:
        pass


class CheckedBumpAllocator(Heap):
    """A simple bump allocator heap that uses a shadow memory to check reads and write. It will also check for UAF bugs."""

    def __init__(
        self, address: int, size: int, guard_bytes: int, *args, **kwargs
    ) -> None:
        super().__init__(address, size, *args, **kwargs)
        self._shadow = [0] * size
        self._current_free_offset = 0
        self._guard_size = guard_bytes
        self._allocations: typing.Dict[int, int] = {}

    def get_used(self) -> int:
        return self._current_free_offset - self.address

    def _is_safe(self, value: state.Value):
        if (
            self.get_used() + value.get_size() + self._guard_size
        ) > self.get_capacity():
            raise ValueError("Memory is full")

    def allocate(self, value: state.Value) -> int:
        self._is_safe(value)
        value_size = value.get_size()

        offset = self._current_free_offset
        self._current_free_offset += value_size + self._guard_size

        for i in range(value_size):
            self._shadow[offset + i] = 1

        self._allocations[offset] = value_size
        self[offset] = value
        return self.address + offset

    def free(self, address: int) -> None:
        if address in self._allocations:
            offset = address - self.address
            size = self._allocations[address]
            for i in range(size):
                self._shadow[offset + i] = 2
            del self._allocations[address]
        else:
            raise ValueError(f"Invalid Free at {hex(address)}")

    def check_access(
        self,
        emulator: smallworld.emulators.Emulator,
        address: int,
        size: int,
        content: bytes,
    ):
        offset = address - self.address
        if (offset + size) > size:
            raise ValueError(f"Invalid access at {hex(address)} of size {size}")

        for i in range(size):
            shadow_byte = self._shadow[offset + i]
            if shadow_byte == 1:
                continue
            elif shadow_byte == 0:
                raise ValueError(f"Access uninitialized bytes at {hex(address + i)}")
            elif shadow_byte == 2:
                raise ValueError(f"Access freed memory at {hex(address + i)}")

        return bytes

    def apply(self, emulator: typing.Any) -> None:
        super().apply(emulator)
        emulator.hook_memory_read(
            self.address, self.address + self.size, self.check_access
        )
        emulator.hook_memory_write(
            self.address, self.address + self.size, self.check_access
        )


__all__ = ["Heap", "BumpAllocator", "CheckedBumpAllocator"]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import MIPSSysVModel


class MIPSSysVFclose(Fclose, MIPSSysVModel):
    pass


class MIPSSysVFeof(Feof, MIPSSysVModel):
    pass


class MIPSSysVFerror(Ferror, MIPSSysVModel):
    pass


class MIPSSysVFgetc(Fgetc, MIPSSysVModel):
    pass


class MIPSSysVFgets(Fgets, MIPSSysVModel):
    pass


class MIPSSysVFopen(Fopen, MIPSSysVModel):
    pass


class MIPSSysVFprintf(Fprintf, MIPSSysVModel):
    pass


class MIPSSysVFputc(Fputc, MIPSSysVModel):
    pass


class MIPSSysVFputs(Fputs, MIPSSysVModel):
    pass


class MIPSSysVFread(Fread, MIPSSysVModel):
    pass


class MIPSSysVFscanf(Fscanf, MIPSSysVModel):
    pass


class MIPSSysVFseek(Fseek, MIPSSysVModel):
    pass


class MIPSSysVFtell(Ftell, MIPSSysVModel):
    pass


class MIPSSysVFwrite(Fwrite, MIPSSysVModel):
    pass


class MIPSSysVGetc(Getc, MIPSSysVModel):
    pass


class MIPSSysVGetchar(Getchar, MIPSSysVModel):
    pass


class MIPSSysVGets(Gets, MIPSSysVModel):
    pass


class MIPSSysVPrintf(Printf, MIPSSysVModel):
    pass


class MIPSSysVPutc(Putc, MIPSSysVModel):
    pass


class MIPSSysVPutchar(Putchar, MIPSSysVModel):
    pass


class MIPSSysVPuts(Puts, MIPSSysVModel):
    pass


class MIPSSysVRemove(Remove, MIPSSysVModel):
    pass


class MIPSSysVRename(Rename, MIPSSysVModel):
    pass


class MIPSSysVRewind(Rewind, MIPSSysVModel):
    pass


class MIPSSysVScanf(Scanf, MIPSSysVModel):
    pass


class MIPSSysVSnprintf(Snprintf, MIPSSysVModel):
    pass


class MIPSSysVSprintf(Sprintf, MIPSSysVModel):
    pass


class MIPSSysVSscanf(Sscanf, MIPSSysVModel):
    pass


class MIPSSysVClearerr(Clearerr, MIPSSysVModel):
    pass


class MIPSSysVFflush(Fflush, MIPSSysVModel):
    pass


class MIPSSysVFreopen(Freopen, MIPSSysVModel):
    pass


class MIPSSysVFgetpos(Fgetpos, MIPSSysVModel):
    pass


class MIPSSysVFsetpos(Fsetpos, MIPSSysVModel):
    pass


class MIPSSysVTmpfile(Tmpfile, MIPSSysVModel):
    pass


class MIPSSysVTmpnam(Tmpnam, MIPSSysVModel):
    pass


class MIPSSysVUngetc(Ungetc, MIPSSysVModel):
    pass


class MIPSSysVVfprintf(Vfprintf, MIPSSysVModel):
    pass


class MIPSSysVVfscanf(Vfscanf, MIPSSysVModel):
    pass


class MIPSSysVVprintf(Vprintf, MIPSSysVModel):
    pass


class MIPSSysVVscanf(Vscanf, MIPSSysVModel):
    pass


class MIPSSysVVsnprintf(Vsnprintf, MIPSSysVModel):
    pass


class MIPSSysVVsprintf(Vsprintf, MIPSSysVModel):
    pass


class MIPSSysVVsscanf(Vsscanf, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVFclose",
    "MIPSSysVFeof",
    "MIPSSysVFerror",
    "MIPSSysVFgetc",
    "MIPSSysVFgets",
    "MIPSSysVFopen",
    "MIPSSysVFprintf",
    "MIPSSysVFputc",
    "MIPSSysVFputs",
    "MIPSSysVFread",
    "MIPSSysVFscanf",
    "MIPSSysVFseek",
    "MIPSSysVFtell",
    "MIPSSysVFwrite",
    "MIPSSysVGetc",
    "MIPSSysVGetchar",
    "MIPSSysVGets",
    "MIPSSysVPrintf",
    "MIPSSysVPutc",
    "MIPSSysVPutchar",
    "MIPSSysVPuts",
    "MIPSSysVRemove",
    "MIPSSysVRename",
    "MIPSSysVRewind",
    "MIPSSysVScanf",
    "MIPSSysVSnprintf",
    "MIPSSysVSprintf",
    "MIPSSysVSscanf",
    "MIPSSysVClearerr",
    "MIPSSysVFflush",
    "MIPSSysVFreopen",
    "MIPSSysVFgetpos",
    "MIPSSysVFsetpos",
    "MIPSSysVTmpfile",
    "MIPSSysVTmpnam",
    "MIPSSysVUngetc",
    "MIPSSysVVfprintf",
    "MIPSSysVVfscanf",
    "MIPSSysVVprintf",
    "MIPSSysVVscanf",
    "MIPSSysVVsnprintf",
    "MIPSSysVVsprintf",
    "MIPSSysVVsscanf",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVMemcpy(Memcpy, LoongArch64SysVModel):
    pass


class LoongArch64SysVMemmove(Memmove, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrcpy(Strcpy, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrncpy(Strncpy, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrcat(Strcat, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrncat(Strncat, LoongArch64SysVModel):
    pass


class LoongArch64SysVMemcmp(Memcmp, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrncmp(Strncmp, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrcmp(Strcmp, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrcoll(Strcoll, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrxfrm(Strxfrm, LoongArch64SysVModel):
    pass


class LoongArch64SysVMemchr(Memchr, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrchr(Strchr, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrcspn(Strcspn, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrpbrk(Strpbrk, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrrchr(Strrchr, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrspn(Strspn, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrstr(Strstr, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrtok(Strtok, LoongArch64SysVModel):
    pass


class LoongArch64SysVMemset(Memset, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrerror(Strerror, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrlen(Strlen, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVMemcpy",
    "LoongArch64SysVMemmove",
    "LoongArch64SysVStrcpy",
    "LoongArch64SysVStrncpy",
    "LoongArch64SysVStrcat",
    "LoongArch64SysVStrncat",
    "LoongArch64SysVMemcmp",
    "LoongArch64SysVStrncmp",
    "LoongArch64SysVStrcmp",
    "LoongArch64SysVStrcoll",
    "LoongArch64SysVStrxfrm",
    "LoongArch64SysVMemchr",
    "LoongArch64SysVStrchr",
    "LoongArch64SysVStrcspn",
    "LoongArch64SysVStrpbrk",
    "LoongArch64SysVStrrchr",
    "LoongArch64SysVStrspn",
    "LoongArch64SysVStrstr",
    "LoongArch64SysVStrtok",
    "LoongArch64SysVMemset",
    "LoongArch64SysVStrerror",
    "LoongArch64SysVStrlen",
]



from ....c99.signal import Raise, Signal
from ..systemv import RiscV64SysVModel


class RiscV64SysVRaise(Raise, RiscV64SysVModel):
    pass


class RiscV64SysVSignal(Signal, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVRaise",
    "RiscV64SysVSignal",
]



import unicorn

from ....platforms import Architecture, Byteorder
from .machdef import UnicornMachineDef


class AArch64MachineDef(UnicornMachineDef):
    arch = Architecture.AARCH64
    byteorder = Byteorder.LITTLE

    uc_arch = unicorn.UC_ARCH_ARM64
    uc_mode = unicorn.UC_MODE_ARM

    _registers = {
        # *** General Purpose Registers ***
        "x0": unicorn.arm64_const.UC_ARM64_REG_X0,
        "w0": unicorn.arm64_const.UC_ARM64_REG_W0,
        "x1": unicorn.arm64_const.UC_ARM64_REG_X1,
        "w1": unicorn.arm64_const.UC_ARM64_REG_W1,
        "x2": unicorn.arm64_const.UC_ARM64_REG_X2,
        "w2": unicorn.arm64_const.UC_ARM64_REG_W2,
        "x3": unicorn.arm64_const.UC_ARM64_REG_X3,
        "w3": unicorn.arm64_const.UC_ARM64_REG_W3,
        "x4": unicorn.arm64_const.UC_ARM64_REG_X4,
        "w4": unicorn.arm64_const.UC_ARM64_REG_W4,
        "x5": unicorn.arm64_const.UC_ARM64_REG_X5,
        "w5": unicorn.arm64_const.UC_ARM64_REG_W5,
        "x6": unicorn.arm64_const.UC_ARM64_REG_X6,
        "w6": unicorn.arm64_const.UC_ARM64_REG_W6,
        "x7": unicorn.arm64_const.UC_ARM64_REG_X7,
        "w7": unicorn.arm64_const.UC_ARM64_REG_W7,
        "x8": unicorn.arm64_const.UC_ARM64_REG_X8,
        "w8": unicorn.arm64_const.UC_ARM64_REG_W8,
        "x9": unicorn.arm64_const.UC_ARM64_REG_X9,
        "w9": unicorn.arm64_const.UC_ARM64_REG_W9,
        "x10": unicorn.arm64_const.UC_ARM64_REG_X10,
        "w10": unicorn.arm64_const.UC_ARM64_REG_W10,
        "x11": unicorn.arm64_const.UC_ARM64_REG_X11,
        "w11": unicorn.arm64_const.UC_ARM64_REG_W11,
        "x12": unicorn.arm64_const.UC_ARM64_REG_X12,
        "w12": unicorn.arm64_const.UC_ARM64_REG_W12,
        "x13": unicorn.arm64_const.UC_ARM64_REG_X13,
        "w13": unicorn.arm64_const.UC_ARM64_REG_W13,
        "x14": unicorn.arm64_const.UC_ARM64_REG_X14,
        "w14": unicorn.arm64_const.UC_ARM64_REG_W14,
        "x15": unicorn.arm64_const.UC_ARM64_REG_X15,
        "w15": unicorn.arm64_const.UC_ARM64_REG_W15,
        "x16": unicorn.arm64_const.UC_ARM64_REG_X16,
        "w16": unicorn.arm64_const.UC_ARM64_REG_W16,
        "x17": unicorn.arm64_const.UC_ARM64_REG_X17,
        "w17": unicorn.arm64_const.UC_ARM64_REG_W17,
        "x18": unicorn.arm64_const.UC_ARM64_REG_X18,
        "w18": unicorn.arm64_const.UC_ARM64_REG_W18,
        "x19": unicorn.arm64_const.UC_ARM64_REG_X19,
        "w19": unicorn.arm64_const.UC_ARM64_REG_W19,
        "x20": unicorn.arm64_const.UC_ARM64_REG_X20,
        "w20": unicorn.arm64_const.UC_ARM64_REG_W20,
        "x21": unicorn.arm64_const.UC_ARM64_REG_X21,
        "w21": unicorn.arm64_const.UC_ARM64_REG_W21,
        "x22": unicorn.arm64_const.UC_ARM64_REG_X22,
        "w22": unicorn.arm64_const.UC_ARM64_REG_W22,
        "x23": unicorn.arm64_const.UC_ARM64_REG_X23,
        "w23": unicorn.arm64_const.UC_ARM64_REG_W23,
        "x24": unicorn.arm64_const.UC_ARM64_REG_X24,
        "w24": unicorn.arm64_const.UC_ARM64_REG_W24,
        "x25": unicorn.arm64_const.UC_ARM64_REG_X25,
        "w25": unicorn.arm64_const.UC_ARM64_REG_W25,
        "x26": unicorn.arm64_const.UC_ARM64_REG_X26,
        "w26": unicorn.arm64_const.UC_ARM64_REG_W26,
        "x27": unicorn.arm64_const.UC_ARM64_REG_X27,
        "w27": unicorn.arm64_const.UC_ARM64_REG_W27,
        "x28": unicorn.arm64_const.UC_ARM64_REG_X28,
        "w28": unicorn.arm64_const.UC_ARM64_REG_W28,
        "x29": unicorn.arm64_const.UC_ARM64_REG_X29,
        "w29": unicorn.arm64_const.UC_ARM64_REG_W29,
        "x30": unicorn.arm64_const.UC_ARM64_REG_X30,
        "w30": unicorn.arm64_const.UC_ARM64_REG_W30,
        # *** Program Counter ***
        "pc": unicorn.arm64_const.UC_ARM64_REG_PC,
        # *** Stack Pointer ***
        "sp": unicorn.arm64_const.UC_ARM64_REG_SP,
        "wsp": unicorn.arm64_const.UC_ARM64_REG_WSP,
        # *** Frame Pointer ***
        "fp": unicorn.arm64_const.UC_ARM64_REG_FP,
        # *** Link Register ***
        "lr": unicorn.arm64_const.UC_ARM64_REG_LR,
        # *** Zero Register ***
        "xzr": unicorn.arm64_const.UC_ARM64_REG_INVALID,
        "wzr": unicorn.arm64_const.UC_ARM64_REG_INVALID,
        # *** System Control Registers ***
        # NOTE: "_elX" indicates that only exception level X or greater can access this register.
        # NOTE: This list is far from complete; it only covers what Unicorn supports
        # Condition Code Register
        "fpcr": unicorn.arm64_const.UC_ARM64_REG_FPCR,
        # Floating Point Status Register
        "fpsr": unicorn.arm64_const.UC_ARM64_REG_FPSR,
        # Banked stack pointers for exception handlers
        "sp_el0": unicorn.arm64_const.UC_ARM64_REG_SP_EL0,
        "sp_el1": unicorn.arm64_const.UC_ARM64_REG_SP_EL1,
        "sp_el2": unicorn.arm64_const.UC_ARM64_REG_SP_EL2,
        "sp_el3": unicorn.arm64_const.UC_ARM64_REG_SP_EL3,
        # Banked link registers for exception handlers
        # NOTE: Unicorn thinks there's an elr_el0; according to docs, it doesn't exist
        "elr_el1": unicorn.arm64_const.UC_ARM64_REG_ELR_EL1,
        "elr_el2": unicorn.arm64_const.UC_ARM64_REG_ELR_EL2,
        "elr_el3": unicorn.arm64_const.UC_ARM64_REG_ELR_EL3,
        # Banked exception syndrome registers for exception handlers
        # NOTE: Unicorn thinks there's a far_el0; according to docs, it doesn't exist
        "far_el1": unicorn.arm64_const.UC_ARM64_REG_FAR_EL1,
        "far_el2": unicorn.arm64_const.UC_ARM64_REG_FAR_EL2,
        "far_el3": unicorn.arm64_const.UC_ARM64_REG_FAR_EL3,
        # Banked vector base address registers for exception handlers
        # NOTE: vbar_el0 and vbar_el1 are aliases for each other.
        # Since vbar_el0 doesn't exist in angr, vbar_el1 has to be the "real" copy.
        "vbar_el1": unicorn.arm64_const.UC_ARM64_REG_VBAR_EL1,
        "vbar_el0": unicorn.arm64_const.UC_ARM64_REG_VBAR_EL0,
        "vbar_el2": unicorn.arm64_const.UC_ARM64_REG_VBAR_EL2,
        "vbar_el3": unicorn.arm64_const.UC_ARM64_REG_VBAR_EL3,
        # Coprocessor access control register
        "cpacr_el1": unicorn.arm64_const.UC_ARM64_REG_CPACR_EL1,
        # Memory Attribute Indirection Register
        "mair_el1": unicorn.arm64_const.UC_ARM64_REG_MAIR_EL1,
        # Physical Address Register
        "par_el1": unicorn.arm64_const.UC_ARM64_REG_PAR_EL1,
        # Translation Table Zero Base Register
        "ttbr0_el1": unicorn.arm64_const.UC_ARM64_REG_TTBR0_EL1,
        # Translation Table One Base Register
        "ttbr1_el1": unicorn.arm64_const.UC_ARM64_REG_TTBR1_EL1,
        # Thread ID Register
        # NOTE: According to docs, there should be an el2 and el3 copy, too.
        "tpidr_el0": unicorn.arm64_const.UC_ARM64_REG_TPIDR_EL0,
        "tpidr_el1": unicorn.arm64_const.UC_ARM64_REG_TPIDR_EL1,
        # Userspace-visible Thread ID register
        "tpidrro_el0": unicorn.arm64_const.UC_ARM64_REG_TPIDRRO_EL0,
        # *** Floating Point Registers ***
        # Scalar Floating Point Registers
        "q0": unicorn.arm64_const.UC_ARM64_REG_Q0,
        "d0": unicorn.arm64_const.UC_ARM64_REG_D0,
        "s0": unicorn.arm64_const.UC_ARM64_REG_S0,
        "h0": unicorn.arm64_const.UC_ARM64_REG_H0,
        "b0": unicorn.arm64_const.UC_ARM64_REG_B0,
        "q1": unicorn.arm64_const.UC_ARM64_REG_Q1,
        "d1": unicorn.arm64_const.UC_ARM64_REG_D1,
        "s1": unicorn.arm64_const.UC_ARM64_REG_S1,
        "h1": unicorn.arm64_const.UC_ARM64_REG_H1,
        "b1": unicorn.arm64_const.UC_ARM64_REG_B1,
        "q2": unicorn.arm64_const.UC_ARM64_REG_Q2,
        "d2": unicorn.arm64_const.UC_ARM64_REG_D2,
        "s2": unicorn.arm64_const.UC_ARM64_REG_S2,
        "h2": unicorn.arm64_const.UC_ARM64_REG_H2,
        "b2": unicorn.arm64_const.UC_ARM64_REG_B2,
        "q3": unicorn.arm64_const.UC_ARM64_REG_Q3,
        "d3": unicorn.arm64_const.UC_ARM64_REG_D3,
        "s3": unicorn.arm64_const.UC_ARM64_REG_S3,
        "h3": unicorn.arm64_const.UC_ARM64_REG_H3,
        "b3": unicorn.arm64_const.UC_ARM64_REG_B3,
        "q4": unicorn.arm64_const.UC_ARM64_REG_Q4,
        "d4": unicorn.arm64_const.UC_ARM64_REG_D4,
        "s4": unicorn.arm64_const.UC_ARM64_REG_S4,
        "h4": unicorn.arm64_const.UC_ARM64_REG_H4,
        "b4": unicorn.arm64_const.UC_ARM64_REG_B4,
        "q5": unicorn.arm64_const.UC_ARM64_REG_Q5,
        "d5": unicorn.arm64_const.UC_ARM64_REG_D5,
        "s5": unicorn.arm64_const.UC_ARM64_REG_S5,
        "h5": unicorn.arm64_const.UC_ARM64_REG_H5,
        "b5": unicorn.arm64_const.UC_ARM64_REG_B5,
        "q6": unicorn.arm64_const.UC_ARM64_REG_Q6,
        "d6": unicorn.arm64_const.UC_ARM64_REG_D6,
        "s6": unicorn.arm64_const.UC_ARM64_REG_S6,
        "h6": unicorn.arm64_const.UC_ARM64_REG_H6,
        "b6": unicorn.arm64_const.UC_ARM64_REG_B6,
        "q7": unicorn.arm64_const.UC_ARM64_REG_Q7,
        "d7": unicorn.arm64_const.UC_ARM64_REG_D7,
        "s7": unicorn.arm64_const.UC_ARM64_REG_S7,
        "h7": unicorn.arm64_const.UC_ARM64_REG_H7,
        "b7": unicorn.arm64_const.UC_ARM64_REG_B7,
        "q8": unicorn.arm64_const.UC_ARM64_REG_Q8,
        "d8": unicorn.arm64_const.UC_ARM64_REG_D8,
        "s8": unicorn.arm64_const.UC_ARM64_REG_S8,
        "h8": unicorn.arm64_const.UC_ARM64_REG_H8,
        "b8": unicorn.arm64_const.UC_ARM64_REG_B8,
        "q9": unicorn.arm64_const.UC_ARM64_REG_Q9,
        "d9": unicorn.arm64_const.UC_ARM64_REG_D9,
        "s9": unicorn.arm64_const.UC_ARM64_REG_S9,
        "h9": unicorn.arm64_const.UC_ARM64_REG_H9,
        "b9": unicorn.arm64_const.UC_ARM64_REG_B9,
        "q10": unicorn.arm64_const.UC_ARM64_REG_Q10,
        "d10": unicorn.arm64_const.UC_ARM64_REG_D10,
        "s10": unicorn.arm64_const.UC_ARM64_REG_S10,
        "h10": unicorn.arm64_const.UC_ARM64_REG_H10,
        "b10": unicorn.arm64_const.UC_ARM64_REG_B10,
        "q11": unicorn.arm64_const.UC_ARM64_REG_Q11,
        "d11": unicorn.arm64_const.UC_ARM64_REG_D11,
        "s11": unicorn.arm64_const.UC_ARM64_REG_S11,
        "h11": unicorn.arm64_const.UC_ARM64_REG_H11,
        "b11": unicorn.arm64_const.UC_ARM64_REG_B11,
        "q12": unicorn.arm64_const.UC_ARM64_REG_Q12,
        "d12": unicorn.arm64_const.UC_ARM64_REG_D12,
        "s12": unicorn.arm64_const.UC_ARM64_REG_S12,
        "h12": unicorn.arm64_const.UC_ARM64_REG_H12,
        "b12": unicorn.arm64_const.UC_ARM64_REG_B12,
        "q13": unicorn.arm64_const.UC_ARM64_REG_Q13,
        "d13": unicorn.arm64_const.UC_ARM64_REG_D13,
        "s13": unicorn.arm64_const.UC_ARM64_REG_S13,
        "h13": unicorn.arm64_const.UC_ARM64_REG_H13,
        "b13": unicorn.arm64_const.UC_ARM64_REG_B13,
        "q14": unicorn.arm64_const.UC_ARM64_REG_Q14,
        "d14": unicorn.arm64_const.UC_ARM64_REG_D14,
        "s14": unicorn.arm64_const.UC_ARM64_REG_S14,
        "h14": unicorn.arm64_const.UC_ARM64_REG_H14,
        "b14": unicorn.arm64_const.UC_ARM64_REG_B14,
        "q15": unicorn.arm64_const.UC_ARM64_REG_Q15,
        "d15": unicorn.arm64_const.UC_ARM64_REG_D15,
        "s15": unicorn.arm64_const.UC_ARM64_REG_S15,
        "h15": unicorn.arm64_const.UC_ARM64_REG_H15,
        "b15": unicorn.arm64_const.UC_ARM64_REG_B15,
        "q16": unicorn.arm64_const.UC_ARM64_REG_Q16,
        "d16": unicorn.arm64_const.UC_ARM64_REG_D16,
        "s16": unicorn.arm64_const.UC_ARM64_REG_S16,
        "h16": unicorn.arm64_const.UC_ARM64_REG_H16,
        "b16": unicorn.arm64_const.UC_ARM64_REG_B16,
        "q17": unicorn.arm64_const.UC_ARM64_REG_Q17,
        "d17": unicorn.arm64_const.UC_ARM64_REG_D17,
        "s17": unicorn.arm64_const.UC_ARM64_REG_S17,
        "h17": unicorn.arm64_const.UC_ARM64_REG_H17,
        "b17": unicorn.arm64_const.UC_ARM64_REG_B17,
        "q18": unicorn.arm64_const.UC_ARM64_REG_Q18,
        "d18": unicorn.arm64_const.UC_ARM64_REG_D18,
        "s18": unicorn.arm64_const.UC_ARM64_REG_S18,
        "h18": unicorn.arm64_const.UC_ARM64_REG_H18,
        "b18": unicorn.arm64_const.UC_ARM64_REG_B18,
        "q19": unicorn.arm64_const.UC_ARM64_REG_Q19,
        "d19": unicorn.arm64_const.UC_ARM64_REG_D19,
        "s19": unicorn.arm64_const.UC_ARM64_REG_S19,
        "h19": unicorn.arm64_const.UC_ARM64_REG_H19,
        "b19": unicorn.arm64_const.UC_ARM64_REG_B19,
        "q20": unicorn.arm64_const.UC_ARM64_REG_Q20,
        "d20": unicorn.arm64_const.UC_ARM64_REG_D20,
        "s20": unicorn.arm64_const.UC_ARM64_REG_S20,
        "h20": unicorn.arm64_const.UC_ARM64_REG_H20,
        "b20": unicorn.arm64_const.UC_ARM64_REG_B20,
        "q21": unicorn.arm64_const.UC_ARM64_REG_Q21,
        "d21": unicorn.arm64_const.UC_ARM64_REG_D21,
        "s21": unicorn.arm64_const.UC_ARM64_REG_S21,
        "h21": unicorn.arm64_const.UC_ARM64_REG_H21,
        "b21": unicorn.arm64_const.UC_ARM64_REG_B21,
        "q22": unicorn.arm64_const.UC_ARM64_REG_Q22,
        "d22": unicorn.arm64_const.UC_ARM64_REG_D22,
        "s22": unicorn.arm64_const.UC_ARM64_REG_S22,
        "h22": unicorn.arm64_const.UC_ARM64_REG_H22,
        "b22": unicorn.arm64_const.UC_ARM64_REG_B22,
        "q23": unicorn.arm64_const.UC_ARM64_REG_Q23,
        "d23": unicorn.arm64_const.UC_ARM64_REG_D23,
        "s23": unicorn.arm64_const.UC_ARM64_REG_S23,
        "h23": unicorn.arm64_const.UC_ARM64_REG_H23,
        "b23": unicorn.arm64_const.UC_ARM64_REG_B23,
        "q24": unicorn.arm64_const.UC_ARM64_REG_Q24,
        "d24": unicorn.arm64_const.UC_ARM64_REG_D24,
        "s24": unicorn.arm64_const.UC_ARM64_REG_S24,
        "h24": unicorn.arm64_const.UC_ARM64_REG_H24,
        "b24": unicorn.arm64_const.UC_ARM64_REG_B24,
        "q25": unicorn.arm64_const.UC_ARM64_REG_Q25,
        "d25": unicorn.arm64_const.UC_ARM64_REG_D25,
        "s25": unicorn.arm64_const.UC_ARM64_REG_S25,
        "h25": unicorn.arm64_const.UC_ARM64_REG_H25,
        "b25": unicorn.arm64_const.UC_ARM64_REG_B25,
        "q26": unicorn.arm64_const.UC_ARM64_REG_Q26,
        "d26": unicorn.arm64_const.UC_ARM64_REG_D26,
        "s26": unicorn.arm64_const.UC_ARM64_REG_S26,
        "h26": unicorn.arm64_const.UC_ARM64_REG_H26,
        "b26": unicorn.arm64_const.UC_ARM64_REG_B26,
        "q27": unicorn.arm64_const.UC_ARM64_REG_Q27,
        "d27": unicorn.arm64_const.UC_ARM64_REG_D27,
        "s27": unicorn.arm64_const.UC_ARM64_REG_S27,
        "h27": unicorn.arm64_const.UC_ARM64_REG_H27,
        "b27": unicorn.arm64_const.UC_ARM64_REG_B27,
        "q28": unicorn.arm64_const.UC_ARM64_REG_Q28,
        "d28": unicorn.arm64_const.UC_ARM64_REG_D28,
        "s28": unicorn.arm64_const.UC_ARM64_REG_S28,
        "h28": unicorn.arm64_const.UC_ARM64_REG_H28,
        "b28": unicorn.arm64_const.UC_ARM64_REG_B28,
        "q29": unicorn.arm64_const.UC_ARM64_REG_Q29,
        "d29": unicorn.arm64_const.UC_ARM64_REG_D29,
        "s29": unicorn.arm64_const.UC_ARM64_REG_S29,
        "h29": unicorn.arm64_const.UC_ARM64_REG_H29,
        "b29": unicorn.arm64_const.UC_ARM64_REG_B29,
        "q30": unicorn.arm64_const.UC_ARM64_REG_Q30,
        "d30": unicorn.arm64_const.UC_ARM64_REG_D30,
        "s30": unicorn.arm64_const.UC_ARM64_REG_S30,
        "h30": unicorn.arm64_const.UC_ARM64_REG_H30,
        "b30": unicorn.arm64_const.UC_ARM64_REG_B30,
        "q31": unicorn.arm64_const.UC_ARM64_REG_Q31,
        "d31": unicorn.arm64_const.UC_ARM64_REG_D31,
        "s31": unicorn.arm64_const.UC_ARM64_REG_S31,
        "h31": unicorn.arm64_const.UC_ARM64_REG_H31,
        "b31": unicorn.arm64_const.UC_ARM64_REG_B31,
        # Vector registers
        # TODO: Figure out how to model these
    }



import typing

from .... import platforms
from . import stack


class LoongArchStack(stack.DescendingStack):
    """A stack for a LoongArch CPU"""

    def get_pointer(self) -> int:
        return (self.address + self.size) - self.get_used()

    def get_alignment(self) -> int:
        return 4

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for MIPS64")


class LoongArch64Stack(LoongArchStack):
    """A stack for a LoongArch 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.LOONGARCH64, platforms.Byteorder.LITTLE
    )



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import RiscV64SysVModel


class RiscV64SysVTime(Time, RiscV64SysVModel):
    pass


class RiscV64SysVLocaltime(Localtime, RiscV64SysVModel):
    pass


class RiscV64SysVGmtime(Gmtime, RiscV64SysVModel):
    pass


class RiscV64SysVCtime(Ctime, RiscV64SysVModel):
    pass


class RiscV64SysVAsctime(Asctime, RiscV64SysVModel):
    pass


class RiscV64SysVStrftime(Strftime, RiscV64SysVModel):
    pass


class RiscV64SysVDifftime(Difftime, RiscV64SysVModel):
    pass


class RiscV64SysVMktime(Mktime, RiscV64SysVModel):
    pass


class RiscV64SysVClock(Clock, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVTime",
    "RiscV64SysVLocaltime",
    "RiscV64SysVGmtime",
    "RiscV64SysVCtime",
    "RiscV64SysVAsctime",
    "RiscV64SysVStrftime",
    "RiscV64SysVDifftime",
    "RiscV64SysVMktime",
    "RiscV64SysVClock",
]



from ....platforms import Architecture, Byteorder
from .machdef import PandaMachineDef


class ARMMachineDef(PandaMachineDef):
    panda_arch = "arm"

    # I'm going to define all the ones we are making possible as of now
    # I need to submit a PR to change to X86 32 bit and to includ eflags
    def __init__(self):
        self._registers = {
            "r0",
            "r1",
            "r2",
            "r3",
            "r4",
            "r5",
            "r6",
            "r7",
            "r8",
            "r9",
            "r10",
            "r11",
            "r12",
            "sp",
            "lr",
            "pc",
        }

        self._registers = {i: i for i in self._registers}
        self._registers = self._registers | {
            "sb": "r9",
            "sl": "r10",
            "fp": "r11",
            "ip": "r12",
        }


class ARMMachineMixinM:
    """Mixin for ARM M-series machine models"""

    def __init__(self):
        super().__init__()
        self._registers_m = {
            # NOTE: None of the expected privileged registers exist
            "psr",
            "primask",
            "basepri",
            "faultmask",
            "control",
            "msp",
            "psp",
        }
        self._registers = self._registers | {i: None for i in self._registers_m}


class ARMMachineMixinFP:
    def __init__(self):
        super().__init__()
        self._registers_fp = {
            "fpscr",
            "fpexc",
            "fpsid",
            "mvfr0",
            "mvfr1",
            "d0",
            "s0",
            "s1",
            "d1",
            "s2",
            "s3",
            "d2",
            "s4",
            "s5",
            "d3",
            "s6",
            "s7",
            "d4",
            "s8",
            "s9",
            "d5",
            "s10",
            "s11",
            "d6",
            "s12",
            "s13",
            "d7",
            "s14",
            "s15",
            "d8",
            "s16",
            "s17",
            "d9",
            "s18",
            "s19",
            "d10",
            "s20",
            "s21",
            "d11",
            "s22",
            "s23",
            "d12",
            "s24",
            "s25",
            "d13",
            "s26",
            "s27",
            "d14",
            "s28",
            "s29",
            "d15",
            "s30",
            "s31",
        }
        self._registers = self._registers | {i: None for i in self._registers_fp}


class ARMMachineMixinA:
    """Mixin for ARM A-series machine models"""

    def __init__(self):
        super().__init__()
        # TODO: QEMU doesn't quite support what I expect.
        # I expected to see cpsr and spsr.
        # I either got the CPU model wrong, or something else is weird.
        # (I strongly suspect Panda exposes the bitmasked aliases of cpsr)
        self._registers_a = {
            "cpsr": None,
            "spsr": None,
            "sp_usr": None,
            "lr_usr": None,
            "r8_usr": None,
            "r9_usr": None,
            "r10_usr": None,
            "r11_usr": None,
            "r12_usr": None,
            "sp_hyp": None,
            "spsr_hyp": None,
            "elr_hyp": None,
            "sp_svc": None,
            "lr_svc": None,
            "spsr_svc": None,
            "sp_abt": None,
            "lr_abt": None,
            "spsr_abt": None,
            "sp_und": None,
            "lr_und": None,
            "spsr_und": None,
            "sp_mon": None,
            "lr_mon": None,
            "spsr_mon": None,
            "sp_irq": None,
            "lr_irq": None,
            "spsr_irq": None,
            "sp_fiq": None,
            "lr_fiq": None,
            "spsr_fiq": None,
            "r8_fiq": None,
            "r9_fiq": None,
            "r10_fiq": None,
            "r11_fiq": None,
            "r12_fiq": None,
        }
        self._registers = self._registers | {k: v for k, v in self._registers_a.items()}


class ARMMachineMixinVFP:
    def __init__(self):
        super().__init__()
        self._registers_vfp = {
            # *** Floating-point Control Registers ***
            # Floating-point Status and Control Register
            "fpscr",
            # Floating-point Exception Control Register
            "fpexc",
            # Floating-point System ID Register
            "fpsid",
            # Media and VFP Feature Register 0
            "mvfr0",
            # Media and VFP Feature Register 1
            "mvfr1",
            # *** Floating-point Registers ****
            "q0",
            "d0",
            "s0",
            "s1",
            "d1",
            "s2",
            "s3",
            "q1",
            "d2",
            "s4",
            "s5",
            "d3",
            "s6",
            "s7",
            "q2",
            "d4",
            "s8",
            "s9",
            "d5",
            "s10",
            "s11",
            "q3",
            "d6",
            "s12",
            "s13",
            "d7",
            "s14",
            "s15",
            "q4",
            "d8",
            "s16",
            "s17",
            "d9",
            "s18",
            "s19",
            "q5",
            "d10",
            "s20",
            "s21",
            "d11",
            "s22",
            "s23",
            "q6",
            "d12",
            "s24",
            "s25",
            "d13",
            "s26",
            "s27",
            "q7",
            "d14",
            "s28",
            "s29",
            "d15",
            "s30",
            "s31",
            "q8",
            "d16",
            "d17",
            "q9",
            "d18",
            "d19",
            "q10",
            "d20",
            "d21",
            "q11",
            "d22",
            "d23",
            "q12",
            "d24",
            "d25",
            "q13",
            "d26",
            "d27",
            "q14",
            "d28",
            "d29",
            "q15",
            "d30",
            "d31",
        }
        self._registers = self._registers | {i: None for i in self._registers_vfp}


class ARMv5TMachineDef(ARMMachineMixinM, ARMMachineDef):
    arch = Architecture.ARM_V5T
    byteorder = Byteorder.LITTLE
    cpu = "pxa255"


# TODO: Something's very weird with Panda's Arm 7 models.
# cortex-a9 should be an A-series, but it looks more like an M-series.
# cortex-m4 looks like an M-series, but aborts; I suspect we're missing configuration.
class ARMv7AMachineDef(ARMMachineMixinVFP, ARMMachineMixinA, ARMMachineDef):
    arch = Architecture.ARM_V7A
    byteorder = Byteorder.LITTLE
    cpu = "cortex-a9"


class ARMv7MMachineDef(ARMMachineMixinFP, ARMMachineMixinM, ARMMachineDef):
    arch = Architecture.ARM_V7M
    byteorder = Byteorder.LITTLE
    cpu = "cortex-m4"



from ....platforms import Architecture, Byteorder
from .machdef import PandaMachineDef


class PowerPCMachineDef(PandaMachineDef):
    byteorder = Byteorder.BIG

    panda_arch = "ppc"

    # I'm going to define all the ones we are making possible as of now
    # I need to submit a PR to change to X86 32 bit and to includ eflags
    _registers_identity = {
        "r0",
        "r2",
        "r3",
        "r4",
        "r5",
        "r6",
        "r7",
        "r8",
        "r9",
        "r10",
        "r11",
        "r12",
        "r13",
        "r14",
        "r15",
        "r16",
        "r17",
        "r18",
        "r19",
        "r20",
        "r21",
        "r22",
        "r23",
        "r24",
        "r25",
        "r26",
        "r27",
        "r28",
        "r29",
        "r30",
        "r31",
        "cr0",
        "cr1",
        "cr2",
        "cr3",
        "cr4",
        "cr5",
        "cr6",
        "cr7",
        "pc",
        "sp",
        "lr",
        "ctr",
    }
    _registers_mapping = {
        "r1": "sp",
        "bp": "r31",
    }
    _registers_unsupported = {
        "f0",
        "f1",
        "f2",
        "f3",
        "f4",
        "f5",
        "f6",
        "f7",
        "f8",
        "f9",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f20",
        "f21",
        "f22",
        "f23",
        "f24",
        "f25",
        "f26",
        "f27",
        "f28",
        "f29",
        "f30",
        "f31",
        "xer",
        "fpscr",
    }
    _registers = {i: j for i, j in _registers_mapping.items()}
    _registers = _registers | {i: i for i in _registers_identity}
    _registers = _registers | {i: None for i in _registers_unsupported}


class PowerPC32MachineDef(PowerPCMachineDef):
    arch = Architecture.POWERPC32
    cpu = "ppc32"



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import MIPS64SysVModel


class MIPS64SysVTime(Time, MIPS64SysVModel):
    pass


class MIPS64SysVLocaltime(Localtime, MIPS64SysVModel):
    pass


class MIPS64SysVGmtime(Gmtime, MIPS64SysVModel):
    pass


class MIPS64SysVCtime(Ctime, MIPS64SysVModel):
    pass


class MIPS64SysVAsctime(Asctime, MIPS64SysVModel):
    pass


class MIPS64SysVStrftime(Strftime, MIPS64SysVModel):
    pass


class MIPS64SysVDifftime(Difftime, MIPS64SysVModel):
    pass


class MIPS64SysVMktime(Mktime, MIPS64SysVModel):
    pass


class MIPS64SysVClock(Clock, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVTime",
    "MIPS64SysVLocaltime",
    "MIPS64SysVGmtime",
    "MIPS64SysVCtime",
    "MIPS64SysVAsctime",
    "MIPS64SysVStrftime",
    "MIPS64SysVDifftime",
    "MIPS64SysVMktime",
    "MIPS64SysVClock",
]



import logging

from ...exceptions import AnalysisError

log = logging.getLogger(__name__)


def reg_name_from_offset(arch, addr: int, size: int):
    """Get a register name from its offset and size in the register file

    This should be straightforward, and it is for standard architectures.

    Architectures supported via pcode have a nasty quirk
    where register down-casts happen before reading the register.
    This means angr can't tell the difference between a real sub-register,
    or a read from part of a larger register.
    This has the effect of selecting incorrect sub-registers,
    or hallucinating non-existent sub-registers.

    This function does a bit of searching to try
    and match the specified address/size to the closest register.

    Arguments:
        arch:   angr architecture object
        addr:   address of register write
        size:   size of register write

    Returns:
        A string containing the register name

    Raises:
        AnalysisException: If no matching register is found

    """
    orig_size = size
    done = False

    if (addr, size) in arch.register_size_names:
        # We have an exact match.
        # Don't bother with other checks.
        done = True
    else:
        # Check if we're accessing the tail of a larger register.
        for r_addr, r_size in arch.register_size_names:
            if addr == r_addr + r_size - size:
                addr = r_addr
                size = r_size
                done = True
                break

    # Hypothesize we're accessing the head of a larger register
    # TODO: This assumes power-of-two-sized registers
    #
    # I know this isn't accurate for some architectures.
    # It also won't be accurate if the sleigh model
    # includes non-power-of-two extractions.
    #
    # TODO: Do any architectures have registers larger than 512 bits?
    # If they do, I'm very sorry.
    while not done and size < 1024:
        if (addr, size) in arch.register_size_names:
            done = True
            break

        if not done:
            size = size << 1

    if not done:
        raise AnalysisError(f"Unknown register for {arch.name}: ({addr}, {orig_size})")
    return arch.register_size_names[(addr, size)]



import enum
import typing

import angr
import pypcode

from ....exceptions import EmulationError
from ....platforms import Architecture, Byteorder
from .machdef import GhidraMachineDef


def handle_nop(irsb, i):
    # This op has no impact on user-facing machine state.
    irsb._ops.pop(i)
    return i


def handle_sigtrap(irsb, i):
    # This op should terminate this block with SIGTRAP
    next_addr = irsb._ops[i - 1].inputs[0].offset + 3
    irsb._ops = irsb._ops[0:i]
    irsb.next = next_addr
    irsb._size = next_addr - irsb.addr
    irsb._instruction_addresses = list(
        filter(lambda x: x < next_addr, irsb._instruction_addresses)
    )
    irsb.jumpkind = "Ijk_SigTRAP"
    return i


def handle_sigill(irsb, i):
    # This op should terminate this block with SIGILL
    next_addr = irsb._ops[i - 1].inputs[0].offset + 3
    irsb._ops = irsb._ops[0:i]
    irsb.next = next_addr
    irsb._size = next_addr - irsb.addr
    irsb._instruction_addresses = list(
        filter(lambda x: x < next_addr, irsb._instruction_addresses)
    )
    irsb.jumpkind = "Ijk_SigILL"
    return i


def handle_syscall(irsb, i):
    # This op should terminate this block with a syscall
    next_addr = irsb._ops[i - 1].inputs[0].offset + 3
    irsb._ops = irsb._ops[0:i]
    irsb.next = next_addr
    irsb._size = next_addr - irsb.addr
    irsb._instruction_addresses = list(
        filter(lambda x: x < next_addr, irsb._instruction_addresses)
    )
    irsb.jumpkind = "Ijk_Sys_syscall"

    return i


def handle_nope(irsb, i):
    # This op cannot be handled by angr
    userop = XtensaUserOp(irsb._ops[i].inputs[0].offset)
    raise EmulationError(f"Unhandled user op {userop!r}")


def handle_unimpl(irsb, i):
    # I don't know what this op does yet
    userop = XtensaUserOp(irsb._ops[i].inputs[0].offset)
    raise EmulationError(f"Unimplemented user op {userop!r}")


class UpdatedEnumMeta(enum.EnumMeta):
    def __contains__(cls, obj):
        if isinstance(obj, int):
            return obj in cls._value2member_map_
        return enum.EnumMeta.__contains__(enum.EnumMeta, obj)


class XtensaUserOp(enum.IntEnum, metaclass=UpdatedEnumMeta):
    def __new__(
        cls, val: int, name: str = "", handler: typing.Any = None, desc: str = ""
    ):
        obj = int.__new__(cls, val)
        obj._value_ = val
        obj.short_name = name
        obj.handler = handler
        obj.description = desc
        return obj

    def __init__(
        self, val: int, name: str = "", handler: typing.Any = None, desc: str = ""
    ):
        self._value_: int = val
        self.short_name: str = name
        self.handler: typing.Any = handler
        self.description: str = desc

    def __repr__(self):
        return f"{hex(self.value)}: {self.short_name} - {self.description}"

    BREAKPOINT = 0x00, "breakpoint", handle_sigtrap, "Breakpoint Instruction"
    DHI = 0x01, "dhi", handle_nop, "Data Cache Hit Invalidate"
    DHU = 0x02, "dhu", handle_nop, "Data Cache Hit Unlock"
    DHWB = 0x03, "dhwb", handle_nop, "Data Cache Hit Writeback"
    DHWBI = 0x04, "dhwbi", handle_nop, "Data Cache Hit Writeback Invalidate"
    DII = 0x05, "dii", handle_nop, "Data Cache Index Invalidate"
    DIU = 0x06, "diu", handle_nop, "Data Cache Index Unlock"
    DIWB = 0x07, "diwb", handle_nop, "Data Cache Index Writeback"
    DIWBI = 0x08, "diwbi", handle_nop, "Data Cache Index Writeback Invalidate"
    DPFL = 0x09, "dpfl", handle_nop, "Data Cache Prefetch And Lock"
    DPFR = 0x0A, "dpfr", handle_nop, "Data Cache Prefetch For Read"
    DPFRO = 0x0B, "dpfro", handle_nop, "Data Cache Prefetch For Read Once"
    DPFW = 0x0C, "dpfw", handle_nop, "Data Cache Prefetch For Write"
    DPFWO = 0x0D, "dpfwo", handle_nop, "Data Cache Prefetch For Write Once"
    DSYNC = 0x0E, "dsync", handle_nop, "Load/Store Synchronize"
    ESYNC = 0x0F, "esync", handle_nop, "Execute Synchronize"
    EXCW = 0x10, "excw", handle_sigtrap, "Exception Wait"
    EXTW = 0x11, "extw", handle_sigtrap, "External Wait"
    IDTLB = 0x12, "idtlb", handle_nop, "Invalidate Data TLB Entry"
    IHI = 0x13, "ihi", handle_nop, "Instruction Cache Hit Invalidate"
    IHU = 0x14, "ihu", handle_nop, "Instruction Cache Hit Unlock"
    III = 0x15, "iii", handle_nop, "Instruction Cache Index Invalidate"
    IITLB = 0x16, "iitlb", handle_nop, "Invalidate Instruction TLB Entry"
    IIU = 0x17, "iiu", handle_nop, "Instruction Cache Index Unlock"
    ILL = 0x18, "ill", handle_sigill, "Illegal Instruction"
    IPF = 0x19, "ipf", handle_nop, "Instruction Cache Prefetch"
    IPFL = 0x1A, "ipfl", handle_nop, "Instruction Cache Prefetch And Lock"
    ISYNC = 0x1B, "isync", handle_nop, "Instruction Fetch Synchronize"
    ACQUIRE = 0x1C, "acquire", handle_unimpl, "???"
    LDCT = 0x1D, "ldct", handle_nope, "Load Data Cache Tag"
    LICT = 0x1E, "lict", handle_nope, "Load Instruction Cache Tag"
    LICW = 0x1F, "licw", handle_nope, "Load Instruction Cache Word"
    MEMW = 0x20, "memw", handle_sigtrap, "Memory Wait"
    NSA = 0x21, "nsa", handle_unimpl, "Normalization Shift Amount"
    NSAU = 0x22, "nsau", handle_unimpl, "Normalization Shift Amount Unsigned"
    PDTLB = 0x23, "pdtlb", handle_nope, "Probe Data TLB"
    PITLB = 0x24, "pitlb", handle_nope, "Probe Instruction TLB"
    RDTLB0 = 0x25, "rdtlb0", handle_nope, "Read Data TLB Entry Virtual"
    RDTLB1 = 0x26, "rdtlb1", handle_nope, "Read Data TLB Entry Translation"
    RER = 0x27, "rer", handle_nope, "Read External Register"
    RESTORE4 = 0x28, "restore4", handle_unimpl, "???"
    RESTORE8 = 0x29, "restore8", handle_unimpl, "???"
    RESTORE12 = 0x2A, "restore12", handle_unimpl, "???"
    RFDD = 0x2B, "rfdd", handle_nope, "Return from Debug and Dispatch"
    RFDE = 0x2C, "rfde", handle_nope, "Return from Double Exception"
    RFDO = 0x2D, "rfdo", handle_nope, "Return from Debug Operation"
    RFE = 0x2E, "rfe", handle_nope, "Return from Exception"
    RFI = 0x2F, "rfi", handle_nope, "Return from Interrupt"
    RFME = 0x30, "rfme", handle_nope, "Return from Memory Error"
    RFUE = 0x31, "rfue", handle_nope, "Return from User Exception"
    RFWO = 0x32, "rfwo", handle_nope, "Return from Window Overflow"
    RFWU = 0x33, "rfwu", handle_nope, "Return from Window Underflow"
    RITLB0 = 0x34, "ritlb0", handle_nope, "Read Instruction TLB Entry Virtual"
    RITLB1 = 0x35, "ritlb1", handle_nope, "Read Instruction TLB Entry Translation"
    RSIL = 0x36, "rsil", handle_nope, "Read and Set Interrupt Level"
    RSR = 0x37, "rsr", handle_unimpl, "Read Special Register"
    RSYNC = 0x38, "rsync", handle_nop, "Register Read Synchronize"
    RUR = 0x39, "rur", handle_unimpl, "Read User Register"
    S32C1I = 0x3A, "s32c1i", handle_unimpl, "Store 32-bit compare conditional"
    RELEASE = 0x3B, "release", handle_unimpl, "???"
    RESTOREREGWINDOW = 0x3C, "restoreregwindow", handle_unimpl, "???"
    ROTATEREGWINDOW = 0x3D, "rotateregwindow", handle_unimpl, "???"
    SDCT = 0x3E, "sdct", handle_nope, "Store Data Cache Tag"
    SICT = 0x3F, "sict", handle_nope, "Store Instruction Cache Tag"
    SICW = 0x40, "sicw", handle_nope, "Store Instruction Cache Word"
    SIMCALL = 0x41, "simcall", handle_nop, "Simulator Call"
    SYSCALL = 0x42, "syscall", handle_syscall, "System Call"
    SWAP4 = 0x43, "swap4", handle_unimpl, "???"
    SWAP8 = 0x44, "swap8", handle_unimpl, "???"
    SWAP12 = 0x45, "swap12", handle_unimpl, "???"
    WAITI = 0x46, "waiti", handle_sigtrap, "Wait for Interrupt"
    WDTLB = 0x47, "wdtlb", handle_nop, "Write Data TLB Entry"
    WER = 0x48, "wer", handle_nop, "Write External Register"
    WITLB = 0x49, "witlb", handle_nop, "Write Instruction TLB Entry"
    WSR = 0x4A, "wsr", handle_unimpl, "Write Special Register"
    WUR = 0x4B, "wur", handle_unimpl, "Write User Register"
    XSR = 0x4C, "xsr", handle_unimpl, "Exchange Special Register"


# Angr doesn't have a syscall calling convention for xtensa.
# Remedy that situation.
class SimCCXtensaLinuxSyscall(angr.calling_conventions.SimCCSyscall):
    ARG_REGS = ["a6", "a3", "a4", "a5", "a8", "a9"]
    FP_ARG_REGS: typing.List[str] = []
    RETURN_VAL = angr.calling_conventions.SimRegArg("a2", 4)
    RETURN_ADDR = angr.calling_conventions.SimRegArg("ip_at_syscall", 4)
    ARCH = None

    @classmethod
    def _match(cls, arch, args, sp_data):
        # Never match; only occurs durring syscalls
        return False

    @staticmethod
    def syscall_num(state):
        return state.regs.a2


angr.calling_conventions.register_syscall_cc(
    "Xtensa:LE:32:default", "default", SimCCXtensaLinuxSyscall
)


class XTensaMachineDef(GhidraMachineDef):
    arch = Architecture.XTENSA
    _registers = {f"a{i}": f"a{i}" for i in range(0, 16)} | {
        "pc": "pc",
        "sar": "sar",
        "sp": "a1",
    }

    def successors(self, state: angr.SimState, **kwargs) -> typing.Any:
        # xtensa includes a _LOT_ of custom pcode operations.

        # Fetch or compute the IR block for our state
        if "irsb" in kwargs and kwargs["irsb"] is not None:
            # Someone's already specified an IR block.
            irsb = kwargs["irsb"]
        else:
            # Disable optimization; it doesn't work
            kwargs["opt_level"] = 0

            # Compute the block from the state.
            # Pray to the Powers that kwargs are compatible.
            irsb = state.block(**kwargs).vex

        i = 0
        while i < len(irsb._ops):
            op = irsb._ops[i]
            if op.opcode == pypcode.OpCode.CALLOTHER:
                # This is a user-defined Pcode op.
                # Alter irsb to mimic its behavior, if we can.
                opnum = op.inputs[0].offset

                if opnum not in XtensaUserOp:
                    # Not a userop.
                    raise EmulationError(f"Undefined user op {hex(opnum)}")
                # get the enum struct
                userop = XtensaUserOp(opnum)

                # Invoke the handler
                i = userop.handler(irsb, i)
            else:
                i += 1

        # Force the engine to use our IR block
        kwargs["irsb"] = irsb

        # Turn the crank on the engine
        return super().successors(state, **kwargs)


class XTensaELMachineDef(XTensaMachineDef):
    byteorder = Byteorder.LITTLE
    pcode_language = "Xtensa:LE:32:default"


class XTensaBEMachineDef(XTensaMachineDef):
    byteorder = Byteorder.BIG
    pcode_language = "Xtensa:BE:32:default"



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVBsdSignal(BsdSignal, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVKill(Kill, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVKillpg(Killpg, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPthreadKill(PthreadKill, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPthreadSigmask(PthreadSigmask, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigaction(Sigaction, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigaddset(Sigaddset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigaltstack(Sigaltstack, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigdelset(Sigdelset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigemptyset(Sigemptyset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigfillset(Sigfillset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSighold(Sighold, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigignore(Sigignore, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSiginterrupt(Siginterrupt, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigismember(Sigismember, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigpause(Sigpause, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigpending(Sigpending, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigprocmask(Sigprocmask, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigqueue(Sigqueue, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigrelse(Sigrelse, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigset(Sigset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigsuspend(Sigsuspend, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigtimedwait(Sigtimedwait, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigwait(Sigwait, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSigwaitinfo(Sigwaitinfo, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVBsdSignal",
    "MIPS64ELSysVKill",
    "MIPS64ELSysVKillpg",
    "MIPS64ELSysVPthreadKill",
    "MIPS64ELSysVPthreadSigmask",
    "MIPS64ELSysVSigaction",
    "MIPS64ELSysVSigaddset",
    "MIPS64ELSysVSigaltstack",
    "MIPS64ELSysVSigdelset",
    "MIPS64ELSysVSigemptyset",
    "MIPS64ELSysVSigfillset",
    "MIPS64ELSysVSighold",
    "MIPS64ELSysVSigignore",
    "MIPS64ELSysVSiginterrupt",
    "MIPS64ELSysVSigismember",
    "MIPS64ELSysVSigpause",
    "MIPS64ELSysVSigpending",
    "MIPS64ELSysVSigprocmask",
    "MIPS64ELSysVSigqueue",
    "MIPS64ELSysVSigrelse",
    "MIPS64ELSysVSigset",
    "MIPS64ELSysVSigsuspend",
    "MIPS64ELSysVSigtimedwait",
    "MIPS64ELSysVSigwait",
    "MIPS64ELSysVSigwaitinfo",
]



import abc

from ... import platforms, utils
from .. import state


class CPU(state.StatefulSet):
    """A CPU state object."""

    def __deepcopy__(self, memo):
        new_cpu = (type(self))()
        for x in self.__dict__:
            a = self.__getattribute__(x)
            if type(a) is state.Register:
                new_cpu.__getattribute__(x).set(self.__getattribute__(x).get())
                new_cpu.__getattribute__(x).set_label(
                    self.__getattribute__(x).get_label()
                )
                new_cpu.__getattribute__(x).set_type(
                    self.__getattribute__(x).get_type()
                )
        return new_cpu

    @property
    @abc.abstractmethod
    def platform(self) -> platforms.Platform:
        pass

    @classmethod
    def get_platform(cls) -> platforms.Platform:
        """Get the platform object for this CPU.

        Returns:
            The platform object for this CPU.
        """
        if not isinstance(cls.platform, platforms.Platform):
            raise TypeError(f"{cls.__name__}.platform is not a Platform object")
        return cls.platform

    @classmethod
    def for_platform(cls, platform: platforms.Platform):
        """Get a CPU object for a given platform specifier.

        Arguments:
            platform: The platform specificer for the desired CPU object.

        Returns:
            An instance of the desired CPU.
        """

        try:
            return utils.find_subclass(cls, lambda x: x.get_platform() == platform)
        except ValueError:
            raise ValueError(f"no model for {platform}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.platform})"


__all__ = ["CPU"]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import MIPSELSysVModel


class MIPSELSysVAbs(Abs, MIPSELSysVModel):
    pass


class MIPSELSysVLAbs(LAbs, MIPSELSysVModel):
    pass


class MIPSELSysVLLAbs(LLAbs, MIPSELSysVModel):
    pass


class MIPSELSysVAtof(Atof, MIPSELSysVModel):
    pass


class MIPSELSysVAtoi(Atoi, MIPSELSysVModel):
    pass


class MIPSELSysVAtol(Atol, MIPSELSysVModel):
    pass


class MIPSELSysVAtoll(Atoll, MIPSELSysVModel):
    pass


class MIPSELSysVCalloc(Calloc, MIPSELSysVModel):
    pass


class MIPSELSysVDiv(Div, MIPSELSysVModel):
    pass


class MIPSELSysVLDiv(LDiv, MIPSELSysVModel):
    pass


class MIPSELSysVLLDiv(LLDiv, MIPSELSysVModel):
    pass


class MIPSELSysVExit(Exit, MIPSELSysVModel):
    pass


class MIPSELSysVFree(Free, MIPSELSysVModel):
    pass


class MIPSELSysVMalloc(Malloc, MIPSELSysVModel):
    pass


class MIPSELSysVQSort(QSort, MIPSELSysVModel):
    pass


class MIPSELSysVRand(Rand, MIPSELSysVModel):
    pass


class MIPSELSysVRealloc(Realloc, MIPSELSysVModel):
    pass


class MIPSELSysVSrand(Srand, MIPSELSysVModel):
    pass


class MIPSELSysVAbort(Abort, MIPSELSysVModel):
    pass


class MIPSELSysVAtexit(Atexit, MIPSELSysVModel):
    pass


class MIPSELSysVGetenv(Getenv, MIPSELSysVModel):
    pass


class MIPSELSysVMblen(Mblen, MIPSELSysVModel):
    pass


class MIPSELSysVMbstowcs(Mbstowcs, MIPSELSysVModel):
    pass


class MIPSELSysVMbtowc(Mbtowc, MIPSELSysVModel):
    pass


class MIPSELSysVSystem(System, MIPSELSysVModel):
    pass


class MIPSELSysVWcstombs(Wcstombs, MIPSELSysVModel):
    pass


class MIPSELSysVWctomb(Wctomb, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVAbs",
    "MIPSELSysVLAbs",
    "MIPSELSysVLLAbs",
    "MIPSELSysVAtof",
    "MIPSELSysVAtoi",
    "MIPSELSysVAtol",
    "MIPSELSysVAtoll",
    "MIPSELSysVCalloc",
    "MIPSELSysVDiv",
    "MIPSELSysVLDiv",
    "MIPSELSysVLLDiv",
    "MIPSELSysVExit",
    "MIPSELSysVFree",
    "MIPSELSysVMalloc",
    "MIPSELSysVQSort",
    "MIPSELSysVRand",
    "MIPSELSysVRealloc",
    "MIPSELSysVSrand",
    "MIPSELSysVAbort",
    "MIPSELSysVAtexit",
    "MIPSELSysVGetenv",
    "MIPSELSysVMblen",
    "MIPSELSysVMbstowcs",
    "MIPSELSysVMbtowc",
    "MIPSELSysVSystem",
    "MIPSELSysVWcstombs",
    "MIPSELSysVWctomb",
]



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import MIPSELSysVModel


class MIPSELSysVTime(Time, MIPSELSysVModel):
    pass


class MIPSELSysVLocaltime(Localtime, MIPSELSysVModel):
    pass


class MIPSELSysVGmtime(Gmtime, MIPSELSysVModel):
    pass


class MIPSELSysVCtime(Ctime, MIPSELSysVModel):
    pass


class MIPSELSysVAsctime(Asctime, MIPSELSysVModel):
    pass


class MIPSELSysVStrftime(Strftime, MIPSELSysVModel):
    pass


class MIPSELSysVDifftime(Difftime, MIPSELSysVModel):
    pass


class MIPSELSysVMktime(Mktime, MIPSELSysVModel):
    pass


class MIPSELSysVClock(Clock, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVTime",
    "MIPSELSysVLocaltime",
    "MIPSELSysVGmtime",
    "MIPSELSysVCtime",
    "MIPSELSysVAsctime",
    "MIPSELSysVStrftime",
    "MIPSELSysVDifftime",
    "MIPSELSysVMktime",
    "MIPSELSysVClock",
]



from __future__ import annotations

import logging
import sys
import threading
import typing
from enum import Enum

import capstone
import claripy
import pandare

from ... import exceptions, platforms, utils
from .. import emulator, hookable
from .machdefs import PandaMachineDef

logger = logging.getLogger(__name__)


class PandaEmulator(
    emulator.Emulator,
    hookable.QInstructionHookable,
    hookable.QFunctionHookable,
    hookable.QMemoryReadHookable,
    hookable.QMemoryWriteHookable,
    hookable.QInterruptHookable,
):
    """An emulator for the Panda emulation engine.

    Arguments:
        arch: Architecture ID string
        mode: Mode ID string
        byteorder: Byteorder

    """

    description = "This is a smallworld class encapsulating the Panda emulator."
    name = "smallworld's-panda"
    version = "0.0"

    PAGE_SIZE = 0x1000

    class ThreadState(Enum):
        SETUP = 1
        RUN = 2
        STEP = 3
        BLOCK = 4
        EXIT = 5

    class PandaThread(threading.Thread):
        # If we want to support both basic block iteration and single step iteration
        # We cannot get pc from state, cpu eip only holds the basic block pc
        # so we have to transfer eip, we just need to pass it around
        # do we want to support this?
        # we could run in single step mode always? what are the implications of this?
        # you have to tell me IN ADVANCE when you setup whether you want to step or block
        # through things...
        # other things that are a bit annoying...cb_insn_translate without single step
        # iterates through a "translation" of a  whole "block" of instructions before it
        # starts executing them, whereas single step would not do that; im not even entirely
        # sure at this point how its determining the bounds of a block but it definitely has
        # something to do with mapped memory

        # NOTE: if there is ANY error in the thread panda code (typos) it will just die...
        # be careful in callbacks
        # If we want to support repeated panda instances we need to make this a subprocess, not thread

        def __init__(self, manager, thread_state):
            super().__init__(daemon=True)
            self.manager = manager
            self.platdef = platforms.PlatformDef.for_platform(self.manager.platform)
            self.machdef = PandaMachineDef.for_platform(self.manager.platform)
            self.state = thread_state
            self.panda = None
            self.hook_return = None

        # Functions to update state, this prob should be changed
        def setup_state(self, cpu):
            self.manager.cpu = cpu

        def update_state(self, cpu, pc):
            self.manager.cpu = cpu
            self.manager.pc = pc

        def get_panda_args_from_machdef(self):
            panda_args = []

            if hasattr(self.machdef, "machine"):
                panda_args.extend(["-M", self.machdef.machine])
            else:
                panda_args.extend(["-M", "configurable"])

            if hasattr(self.machdef, "cpu"):  # != "":
                panda_args.extend(["-cpu", self.machdef.cpu])

            panda_args.extend(["-nographic"])
            # At some point we can send something in that only supports singlestep?
            # panda_args.extend(["singlestep"])
            return panda_args

        def run(self):
            panda_args = self.get_panda_args_from_machdef()

            self.panda = pandare.Panda(self.machdef.panda_arch, extra_args=panda_args)

            @self.panda.cb_after_machine_init
            def setup(cpu):
                logger.debug("Panda: setting up state")
                self.setup_state(cpu)
                self.signal_and_wait()

            @self.panda.cb_insn_translate
            def should_run_on_insn(env, pc):
                return True

            @self.panda.cb_insn_exec
            def on_insn(cpu, pc):
                # PowerPC pc move pc to end of instr
                # so we need to do some stuff to fix that
                if self.machdef.panda_arch == "ppc":
                    pc = pc - 4  # DONT BLAME ME, BLAME ALEX H AND ME :)
                self.update_state(cpu, pc)

                if pc in self.manager._exit_points:
                    # stay here until i say die
                    logger.debug("\ton_insn: exit")
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait()
                elif self.state == PandaEmulator.ThreadState.RUN:
                    # keep going until the end
                    logger.debug("\ton_insn: run")
                elif self.state == PandaEmulator.ThreadState.STEP:
                    # stop and wait for me
                    logger.debug("\ton_insn: step")
                    self.signal_and_wait()
                elif self.state == PandaEmulator.ThreadState.BLOCK:
                    # keep going until the end
                    logger.debug("\ton_insn: block")

                logger.debug(f"Panda: on_insn: {hex(pc)}, {self.state}")
                # Check if our pc is in bounds; if not stop
                if (
                    not self.manager._bounds.is_empty()
                    and not self.manager._bounds.contains_value(pc)
                ):
                    logger.debug(f"Panda: {pc} out of bounds")
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait()

                # Always call hooked code first
                if self.manager.all_instructions_hook:
                    self.manager.all_instructions_hook(self.manager)

                if cb := self.manager.is_instruction_hooked(pc):
                    try:
                        cb(self.manager)
                    except exceptions.EmulationStop:
                        self.state = PandaEmulator.ThreadState.EXIT
                        self.signal_and_wait()
                    except Exception as e:
                        logger.exception(
                            f"Exception running instruction hook at {pc:x}"
                        )
                        self.state = PandaEmulator.ThreadState.EXIT
                        self.signal_and_wait(exception=e)

                if cb := self.manager.is_function_hooked(pc):
                    try:
                        logger.info(f"Calling function callback at {hex(pc)}")
                        cb(self.manager)
                        logger.info(f"Completed function callback at {hex(pc)}")
                    except exceptions.EmulationStop:
                        self.state = PandaEmulator.ThreadState.EXIT
                        self.signal_and_wait()
                    except Exception as e:
                        logger.exception(f"Exception running function hook at {pc:x}")
                        self.state = PandaEmulator.ThreadState.EXIT
                        self.signal_and_wait(exception=e)

                    # Mimic a platform-specific "return" instruction.
                    if (
                        self.manager.platform.architecture
                        == platforms.Architecture.X86_32
                    ):
                        # i386: pop a 4-byte value off the stack
                        sp = self.manager.read_register("esp")
                        ret = int.from_bytes(
                            self.manager.read_memory(sp, 4),
                            self.manager.platform.byteorder.value,
                        )
                        self.manager.write_register("esp", sp + 4)
                    elif (
                        self.manager.platform.architecture
                        == platforms.Architecture.X86_64
                    ):
                        # amd64: pop an 8-byte value off the stack
                        sp = self.manager.read_register("rsp")
                        ret = int.from_bytes(
                            self.manager.read_memory(sp, 8),
                            self.manager.platform.byteorder.value,
                        )
                        self.manager.write_register("rsp", sp + 8)
                    elif (
                        self.manager.platform.architecture
                        == platforms.Architecture.AARCH64
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V5T
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V6M
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V6M_THUMB
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V7A
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V7M
                        or self.manager.platform.architecture
                        == platforms.Architecture.ARM_V7R
                        or self.manager.platform.architecture
                        == platforms.Architecture.POWERPC32
                        or self.manager.platform.architecture
                        == platforms.Architecture.POWERPC64
                    ):
                        # aarch64, arm32, powerpc and powerpc64: branch to register 'lr'
                        ret = self.manager.read_register("lr")
                    elif (
                        self.manager.platform.architecture
                        == platforms.Architecture.MIPS32
                        or self.manager.platform.architecture
                        == platforms.Architecture.MIPS64
                    ):
                        # mips32 and mips64: branch to register 'ra'
                        ret = self.manager.read_register("ra")
                    else:
                        logger.error(
                            "Don't know how to return for {self.manager.platform.architecture}"
                        )
                        self.state = PandaEmulator.ThreadState.EXIT
                        self.signal_and_wait()

                    self.manager.write_register("pc", ret)

                # Now, if we for some reason have a different pc
                # then the one that is set for us, break out of this
                # This would be from changing eip in a hook
                # print(f"Panda: {pc}, {self.manager.pc}")
                # print(self.manager.read_register('pc'))
                # if self.manager.read_register("pc") != pc:
                if self.manager.pc != pc:
                    self.panda.libpanda.cpu_loop_exit_noexc(cpu)

                if not self.manager.current_instruction():
                    # report error if function hooking is enabled?
                    pass
                logger.debug(f"\t{self.manager.current_instruction()}")
                self.hook_return = pc + self.manager.current_instruction().size

                return True

            # Used for stepping over blocks
            @self.panda.cb_start_block_exec(enabled=True)
            def on_block(cpu, tb):
                self.update_state(cpu, tb.pc)
                if (
                    self.state == PandaEmulator.ThreadState.BLOCK
                    or self.state == PandaEmulator.ThreadState.SETUP
                ):
                    logger.debug(f"Panda: on_block: {tb}, {self.state}")
                    # We need to pause on the next block and wait
                    self.signal_and_wait()

            # Used for hooking mem reads
            @self.panda.cb_virt_mem_before_read(enabled=True)
            def on_read(cpu, pc, addr, size):
                logger.debug(f"\ton_read: {addr}")
                orig_data = self.panda.virtual_memory_read(self.manager.cpu, addr, size)
                try:
                    if self.manager.all_reads_hook:
                        val = self.manager.all_reads_hook(
                            self.manager, addr, size, orig_data
                        )
                        if val:
                            self.manager.write_memory(addr, val)
                            orig_data = val
                    if cb := self.manager.is_memory_read_hooked(addr, size):
                        val = cb(self.manager, addr, size, orig_data)
                        if val:
                            self.manager.write_memory(addr, val)
                except exceptions.EmulationStop:
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait()
                except Exception as e:
                    logger.exception(
                        "Exception running read hook at {pc:x}, from {addr:x}"
                    )
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait(exception=e)

            # Used for hooking mem writes
            @self.panda.cb_virt_mem_before_write(enabled=True)
            def on_write(cpu, pc, addr, size, buf):
                logger.debug(f"\ton_write: {hex(addr)}")
                byte_val = bytes([buf[i] for i in range(size)])
                try:
                    if self.manager.all_writes_hook:
                        self.manager.all_writes_hook(self.manager, addr, size, byte_val)

                    if cb := self.manager.is_memory_write_hooked(addr, size):
                        cb(self.manager, addr, size, byte_val)
                except exceptions.EmulationStop:
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait()
                except Exception as e:
                    logger.exception(
                        "Exception running read hook at {pc:x}, from {addr:x}"
                    )
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait(exception=e)

            @self.panda.cb_before_handle_interrupt(enabled=True)
            def on_interrupt(cpu, intno):
                logger.debug(f"\ton_interrupt: {intno}")
                try:
                    # First if all interrupts are hooked, run that function
                    if self.manager.all_interrupts_hook:
                        self.manager.all_interrupts_hook(self.manager)
                    # Then run interrupt specific function
                    if cb := self.manager.is_interrupt_hooked(intno):
                        cb(self.manager)
                except exceptions.EmulationStop:
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait()
                except Exception as e:
                    logger.exception("Exception running interrupt hook for {intno}")
                    self.state = PandaEmulator.ThreadState.EXIT
                    self.signal_and_wait(exception=e)

            @self.panda.cb_before_handle_exception(enabled=True)
            def on_exception(cpu, exception_index):
                logger.error(
                    f"Panda for help: you are hitting an exception at {exception_index}."
                )
                self.state = PandaEmulator.ThreadState.EXIT
                # Generate an exception so we exit ungracefully
                try:
                    raise exceptions.EmulationError(
                        f"Panda exception {exception_index}"
                    )
                except exceptions.EmulationError as e:
                    self.signal_and_wait(exception=e)

            self.panda.run()

        # This is a blocking call for this thread
        def signal_and_wait(self, exception: typing.Optional[Exception] = None):
            logger.debug("Signaling main to run")
            with self.manager.condition:
                # Save an exception if someone wanted to raise one
                self.manager.exception = exception

                # Signal that we are done and to run main
                self.manager.run_main = True
                self.manager.condition.notify()

                # Wait until main tells us to run panda
                while not self.manager.run_panda:
                    self.manager.condition.wait()

                # Clear the event for the next iteration
                self.manager.run_panda = False

    def __init__(self, platform: platforms.Platform):
        super().__init__(platform=platform)

        self.PAGE_SIZE = 0x1000
        self.platform = platform

        # Emulator variables
        self.mapped_pages = utils.RangeCollection()

        # Thread/Main sync variables
        self.condition = threading.Condition()
        self.run_panda = False
        self.run_main = False
        self.exception: typing.Optional[Exception] = None

        # Thread communication variables
        self.cpu = None
        self.pc: int = 0

        self.panda_thread = self.PandaThread(self, self.ThreadState.SETUP)
        self.panda_thread.start()

        self.disassembler = capstone.Cs(
            self.panda_thread.platdef.capstone_arch,
            self.panda_thread.platdef.capstone_mode,
        )
        self.disassembler.detail = True

        # Wait until panda is up and ready
        with self.condition:
            while not self.run_main:
                self.condition.wait()
            # Clear the event for the next iteration
            self.run_main = False

    def read_register_content(self, name: str) -> int:
        # If we are reading a "pc" reg, refer to the manager's notion of PC
        # QEMU thinks in blocks, so the register contained in self.cpu will be inaccurate.
        if name == "pc" or name == self.panda_thread.machdef.panda_reg("pc"):
            return self.pc

        if not self.panda_thread.machdef.check_panda_reg(name):
            raise exceptions.UnsupportedRegisterError(
                f"Panda doesn't support register {name} for {self.platform}"
            )
        name = self.panda_thread.machdef.panda_reg(name)

        try:
            return self.panda_thread.panda.arch.get_reg(self.cpu, name)
        except:
            raise exceptions.AnalysisError(f"Failed reading {name} (id: {name})")

    def write_register_content(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        if content is None:
            logger.debug(f"ignoring register write to {name} - no value")
            return

        if isinstance(content, claripy.ast.bv.BV):
            raise exceptions.SymbolicValueError(
                "This emulator cannot handle bitvector expressions"
            )

        if name == "pc" or name == self.panda_thread.machdef.panda_reg("pc"):
            # This is my internal pc
            self.pc = content
            self.panda_thread.panda.arch.set_pc(self.cpu, content)
            return

        if not self.panda_thread.machdef.check_panda_reg(name):
            raise exceptions.UnsupportedRegisterError(
                f"Panda doesn't support register {name} for {self.platform}"
            )

        name = self.panda_thread.machdef.panda_reg(name)
        try:
            self.panda_thread.panda.arch.set_reg(self.cpu, name, content)
        except:
            raise exceptions.AnalysisError(f"Failed writing {name} (id: {name})")

        logger.debug(f"set register {name}={content}")

    def read_memory_content(self, address: int, size: int) -> bytes:
        if size > sys.maxsize:
            raise ValueError(f"{size} is too large (max: {sys.maxsize})")

        return self.panda_thread.panda.virtual_memory_read(self.cpu, address, size)

    def map_memory(self, address: int, size: int) -> None:
        def page_down(address):
            return address // self.PAGE_SIZE

        def page_up(address):
            return (address + self.PAGE_SIZE - 1) // self.PAGE_SIZE

        logger.info(
            f"map_memory:asking for mapping at {hex(address)}, size {hex(size)}"
        )
        # Translate an addressi + size to a page range
        if page_down(address) == page_down(address + size):
            region = (page_down(address), page_up(address + size) + 1)
        else:
            region = (page_down(address), page_up(address + size))

        logger.info(f"map_memory: Page range: {region}")

        # Get the missing pages first. Those are the ones we want to map
        missing_range = self.mapped_pages.get_missing_ranges(region)

        # Map in those pages and change the memory mapping
        # Whatever you do map just map a page size or above
        logger.info(f"Mapping memory {missing_range} page(s).")
        for start_page, end_page in missing_range:
            page_size = end_page - start_page
            logger.info(
                f"Mapping at {hex(start_page * self.PAGE_SIZE)} in panda of size {hex(page_size * self.PAGE_SIZE)}"
            )
            self.panda_thread.panda.map_memory(
                f"{start_page * self.PAGE_SIZE}",
                page_size * self.PAGE_SIZE,
                start_page * self.PAGE_SIZE,
            )
        # Make sure we add our new region to our mapped_pages
        self.mapped_pages.add_range(region)

    def get_memory_map(self) -> typing.List[typing.Tuple[int, int]]:
        return list(self.mapped_pages.ranges)

    def write_memory_content(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        # Should we type check, if content isnt bytes mad?
        if content is None:
            raise ValueError(f"{self.__class__.__name__} requires concrete state")

        if isinstance(content, claripy.ast.bv.BV):
            raise exceptions.SymbolicValueError(
                "This emulator cannot handle bitvector expressions"
            )

        if len(content) > sys.maxsize:
            raise ValueError(f"{len(content)} is too large (max: {sys.maxsize})")

        if not len(content):
            raise ValueError("memory write cannot be empty")

        # FIXME: MIPS64's physical memory space already has contents.
        # The upper 2**32 bytes of a MIPS64 device is reserved for MMIO.
        # attempting to store stuff there will almost certainly not do what you want.
        # This would go away if we figured out how to emulate virtual memory.
        if self.platform.architecture == platforms.Architecture.MIPS64:
            if address >= 2**32 or (address + len(content)) >= 2**32:
                logger.error(
                    f"Attempting to write to {hex(address)} - {hex(address + len(content))} on MIPS64"
                )
                logger.error("This strays into reserved MMIO memory; please don't.")
                raise exceptions.EmulationError("Write to MIPS64 MMIO space")

        # self.panda_thread.panda.physical_memory_write(address, content)

        # NOTE: Panda does not handle writing across segment boundaries gracefully
        #
        # If you map 0x1000 - 0x2000, and then map 0x2000 - 0x3000,
        # writing to 0x1000 - 0x3000 will probably segfault the emulator.
        #
        # This is because QEMU may put adjacent guest memory segments
        # in non-adjacent host memory.
        # Panda only performs the guest-to-host address translation
        # on the initial address; it doesn't check
        # that all requested blocks are in the same segment.

        offset = 0
        size = len(content)

        # NOTE: This algorithm takes advantage of the fact that Python slices
        # can return fewer than the requested number of bytes if the upper bound
        # is past the end of the list being sliced.
        if address % self.PAGE_SIZE != 0:
            block_size = address % self.PAGE_SIZE
            self.panda_thread.panda.physical_memory_write(
                address, content[0:block_size]
            )

            offset += block_size
            size -= block_size

        while size > 0:
            block_address = address + offset
            self.panda_thread.panda.physical_memory_write(
                block_address, content[offset : offset + self.PAGE_SIZE]
            )

            offset += self.PAGE_SIZE
            size -= self.PAGE_SIZE

        logger.debug(f"wrote {len(content)} bytes to 0x{address:x}")

    def disassemble(
        self, code: bytes, base: int, count: typing.Optional[int] = None
    ) -> typing.Tuple[typing.List[capstone.CsInsn], str]:
        # TODO: annotate that offsets are relative.
        #
        # We don't know what the base address is at disassembly time - so we
        # just set it to 0. This means relative address arguments aren't
        # correctly calculated - we should annotate that relative arguments are
        # relative e.g., with a "+" or something.

        instructions = self.disassembler.disasm(code, base)

        disassembly = []
        insns = []
        for i, instruction in enumerate(instructions):
            if count is not None and i >= count:
                break
            insns.append(instruction)
            disassembly.append(f"{instruction.mnemonic} {instruction.op_str}")

        return (insns, "\n".join(disassembly))

    def current_instruction(self) -> capstone.CsInsn:
        pc = self.pc
        code = self.read_memory(pc, 15)
        if code is None:
            raise AssertionError("invalid state")
        for i in self.disassembler.disasm(code, pc):
            return i

    def check(self) -> None:
        if len(self._exit_points) == 0:
            raise exceptions.ConfigurationError(
                "at least one exit point must be set, emulation cannot start"
            )
        if self.panda_thread.state == self.ThreadState.EXIT:
            logger.debug("stopping emulation at exit point")
            raise exceptions.EmulationBounds

    def run(self) -> None:
        self.check()
        logger.info(f"starting emulation at {hex(self.pc)}")
        self.panda_thread.state = self.ThreadState.RUN
        self.signal_and_wait()
        logger.info("emulation complete")

    def signal_and_wait(self) -> None:
        logger.debug("Main signaling panda to run")
        with self.condition:
            # Signal that we are done and to run panda
            self.run_panda = True
            self.condition.notify()

            # Wait until main tells us to run panda
            while not self.run_main:
                self.condition.wait()

            # Clear the event for the next iteration
            self.run_main = False

            # If we had an exception, raise it
            if self.exception is not None:
                raise self.exception

    def step_block(self) -> None:
        self.check()
        if self.panda_thread.state == self.ThreadState.SETUP:
            # Move past setup
            self.signal_and_wait()

        pc = self.pc
        code = self.read_memory(pc, 15)  # longest possible instruction
        if code is None:
            assert False, "impossible state"
        (instr, disas) = self.disassemble(code, pc, 1)

        logger.info(f"block step at 0x{pc:x}: {disas}")

        self.panda_thread.state = self.ThreadState.BLOCK
        self.signal_and_wait()

    def step_instruction(self) -> None:
        self.check()

        if (
            self.panda_thread.state == self.ThreadState.SETUP
            or self.panda_thread.state == self.ThreadState.BLOCK
        ):
            # Move past setup or block
            self.panda_thread.state = self.ThreadState.STEP
            self.signal_and_wait()

        self.panda_thread.state = self.ThreadState.STEP

        pc = self.pc
        logger.debug(f"Step: reading register {pc}")

        code = self.read_memory(pc, 15)  # longest possible instruction
        if code is None:
            assert False, "impossible state"
        (instr, disas) = self.disassemble(code, pc, 1)

        logger.info(f"single step at 0x{pc:x}: {disas}")

        # We can run now and wait at next instr;
        self.signal_and_wait()
        return

    def __repr__(self) -> str:
        return f"PandaEmulator(platform={self.platform})"



import typing

from .... import platforms
from . import stack


class AMD64Stack(stack.DescendingStack):
    """A stack for an AMD 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.X86_64, platforms.Byteorder.LITTLE
    )

    def get_alignment(self) -> int:
        return 16

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        s = cls(*args, **kwargs)
        argv_address = []
        total_strings_bytes = 0
        for i, arg in enumerate(argv):
            arg_size = len(arg)
            total_strings_bytes += arg_size
            argv_address.append((i, s.push_bytes(arg, label=f"argv[{i}]")))

        argc = len(argv)
        total_space = (8 * (argc + 2)) + total_strings_bytes
        padding = 16 - (total_space % 16)
        s.push_bytes(bytes(padding), label="stack alignment padding bytes")
        s.push_integer(0, size=8, label="null terminator of argv array")
        for i, addr in reversed(argv_address):
            s.push_integer(addr, size=8, label=f"pointer to argv[{i}]")
        s.push_integer(argc, size=8, label="argc")
        return s



import logging

from ... import emulators, hinting, state
from .. import analysis
from .angr.nwbt import configure_nwbt_plugins, configure_nwbt_strategy
from .angr.utils import print_state

log = logging.getLogger(__name__)


class AngrNWBTAnalysis(analysis.Analysis):
    name = "angr-nwbt"
    description = "Angr-based Not-Written-By-Trace detection and correction"
    version = "0.0.1"

    def __init__(self, *args, initfunc=None, max_steps=500, **kwargs):
        self.steps_left = max_steps
        self.initfunc = initfunc

    def analysis_preinit(self, emu):
        log.info("Registering plugins")
        configure_nwbt_plugins(emu)

    def analysis_init(self, emu):
        configure_nwbt_strategy(emu)
        if self.initfunc is not None:
            self.initfunc(self, emu.entry)

    def run(self, machine: state.Machine):
        cpu = machine.get_cpu()
        emu = emulators.AngrEmulator(
            cpu.platform, preinit=self.analysis_preinit, init=self.analysis_init
        )
        machine.apply(emu)

        # Extract typedef info from the CPU state,
        # and bind it to the machine state
        for item in cpu:
            if isinstance(item, state.Register):
                if item.get_type() is not None:
                    log.debug(f"Applying type for {item}")
                    emu.state.typedefs.bind_register(item.name, item.get_type())
        for item in machine:
            if isinstance(item, state.memory.code.Executable):
                # TODO: Code is also memory, so it should have types...?
                pass
            elif isinstance(item, state.memory.Memory):
                for offset, value in item.items():
                    if value.get_type() is not None:
                        addr = item.address + offset
                        log.debug(f"Applying type for {hex(addr)}")
                        emu.state.typedefs.bind_address(addr, value.get_type())
            else:
                if not isinstance(item, state.cpus.CPU) and item.type is not None:
                    raise NotImplementedError(
                        f"Applying typedef {item.get_type()} of type {type(item)} not implemented"
                    )

        while (self.steps_left is None or self.steps_left > 0) and not self.step(emu):
            if self.steps_left is not None:
                self.steps_left -= 1

    def _report_status(self, emu):
        for st in emu.mgr.active:
            dis = "\r".join(map(str, st.block().disassembly.insns))
            log.debug(f"Active state: {dis}")
        for st in emu.mgr.unconstrained:
            hint = hinting.OutputHint(
                message="State left the program",
                registers=st.registers.create_hint(),
                memory=st.memory.create_hint(),
            )
            self.hinter.send(hint)
        for st in emu.mgr.deadended:
            hint = hinting.OutputHint(
                message="State exited due to breakpoint",
                registers=st.registers.create_hint(),
                memory=st.memory.create_hint(),
            )
            self.hinter.send(hint)
        for st in emu.mgr.unsat:
            hint = hinting.OutputHint(
                message="State cannot continue; constraints unsat",
                registers=st.registers.create_hint(),
                memory=st.memory.create_hint(),
            )
            self.hinter.send(hint)
        for err in emu.mgr.errored:
            print_state(log.info, err.state, "error")
            log.error(
                "\tError:",
                exc_info=(type(err.error), err.error, err.error.__traceback__),
            )
        log.info(f"Summary: {emu.mgr}")

    def step(self, emu):
        # Report our current status
        self._report_status(emu)
        # In our case, return states are unconstrained.
        emu.mgr.move(from_stash="deadended", to_stash="done")
        emu.mgr.move(from_stash="unconstrained", to_stash="done")
        # Drop unsat states once we've logged them.
        emu.mgr.drop(stash="unsat")
        self._report_status(emu)
        return emu.step()



from ....c99.signal import Raise, Signal
from ..systemv import AMD64SysVModel


class AMD64SysVRaise(Raise, AMD64SysVModel):
    pass


class AMD64SysVSignal(Signal, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVRaise",
    "AMD64SysVSignal",
]



from . import code, heap, stack
from .elf import *  # noqa: F401, F403
from .elf import __all__ as __elf__
from .memory import *  # noqa: F401, F403
from .memory import __all__ as __memory__

__all__ = __memory__ + __elf__ + ["stack", "heap", "code"]



import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef

# NOTE: RiscV is designed to be an extensible language.
# The core ISA is incredibly simple,
# with a large number of optional extensions available
# to chip manufacturers
#
# Support for any specific extension depends on the emulator.


class RiscV64(PlatformDef):
    architecture = Architecture.RISCV64
    byteorder = Byteorder.LITTLE

    address_size = 8

    capstone_arch = capstone.CS_ARCH_RISCV
    capstone_mode = capstone.CS_MODE_RISCV64

    conditional_branch_mnemonics = {
        # Core conditional branch instructions
        "beq",
        "bne",
        "blt",
        "bltu",
        "ble",
        "bleu",
        "bgt",
        "bgtu",
        "bge",
        "bgeu",
    }

    # RiscV core has no dedicated compare instructions
    # (they're rather smug about this fact).
    #
    # - The "Zicond" extension adds conditional move instructions.
    # - The "Zacas" extension adds compare-exchange instructions
    compare_mnemonics = set()

    pc_register = "pc"
    sp_register = "sp"

    # Special registers:
    # - x0 is wired to zero
    # - x1 is the link register
    # - x2 is the stack pointer
    # - x3 is the global pointer
    # - x4 is the thread pointer
    general_purpose_registers = [f"x{i}" for i in range(0, 32)]

    registers = {
        # *** General-Purpose Registers ***
        # x0 is wired to 0, and aliased as "zero"
        "x0": RegisterDef(name="x0", size=8),
        "zero": RegisterAliasDef(name="zero", parent="x0", size=8, offset=0),
        # x1 acts as the link register
        # NOTE: ra is the official name; lr might be an angr invention.
        "x1": RegisterDef(name="x1", size=8),
        "ra": RegisterAliasDef(name="ra", parent="x1", size=8, offset=0),
        # x2 acts as the stack pointer
        "x2": RegisterDef(name="x2", size=8),
        "sp": RegisterAliasDef(name="sp", parent="x2", size=8, offset=0),
        # x3 acts as the global pointer
        "x3": RegisterDef(name="x3", size=8),
        "gp": RegisterAliasDef(name="gp", parent="x3", size=8, offset=0),
        # x4 acts as the thread pointer
        "x4": RegisterDef(name="x4", size=8),
        "tp": RegisterAliasDef(name="tp", parent="x4", size=8, offset=0),
        # x5 is a temporary register
        "x5": RegisterDef(name="x5", size=8),
        "t0": RegisterAliasDef(name="t0", parent="x5", size=8, offset=0),
        # x6 is a temporary register
        "x6": RegisterDef(name="x6", size=8),
        "t1": RegisterAliasDef(name="t1", parent="x6", size=8, offset=0),
        # x7 is a temporary register
        "x7": RegisterDef(name="x7", size=8),
        "t2": RegisterAliasDef(name="t2", parent="x7", size=8, offset=0),
        # x8 is a callee-saved register
        "x8": RegisterDef(name="x8", size=8),
        "s0": RegisterAliasDef(name="s0", parent="x8", size=8, offset=0),
        # x9 is a callee-saved register
        "x9": RegisterDef(name="x9", size=8),
        "s1": RegisterAliasDef(name="s1", parent="x9", size=8, offset=0),
        # x10 is argument 0
        "x10": RegisterDef(name="x10", size=8),
        "a0": RegisterAliasDef(name="a0", parent="x10", size=8, offset=0),
        # x11 is argument 1
        "x11": RegisterDef(name="x11", size=8),
        "a1": RegisterAliasDef(name="a1", parent="x11", size=8, offset=0),
        # x12 is argument 2
        "x12": RegisterDef(name="x12", size=8),
        "a2": RegisterAliasDef(name="a2", parent="x12", size=8, offset=0),
        # x13 is argument 3
        "x13": RegisterDef(name="x13", size=8),
        "a3": RegisterAliasDef(name="a3", parent="x13", size=8, offset=0),
        # x14 is argument 4
        "x14": RegisterDef(name="x14", size=8),
        "a4": RegisterAliasDef(name="a4", parent="x14", size=8, offset=0),
        # x15 is argument 5
        "x15": RegisterDef(name="x15", size=8),
        "a5": RegisterAliasDef(name="a5", parent="x15", size=8, offset=0),
        # x16 is argument 6
        "x16": RegisterDef(name="x16", size=8),
        "a6": RegisterAliasDef(name="a6", parent="x16", size=8, offset=0),
        # x17 is argument 7
        "x17": RegisterDef(name="x17", size=8),
        "a7": RegisterAliasDef(name="a7", parent="x17", size=8, offset=0),
        # x18 is a callee-saved register
        "x18": RegisterDef(name="x18", size=8),
        "s2": RegisterAliasDef(name="s2", parent="x18", size=8, offset=0),
        # x19 is a callee-saved register
        "x19": RegisterDef(name="x19", size=8),
        "s3": RegisterAliasDef(name="s3", parent="x19", size=8, offset=0),
        # x20 is a callee-saved register
        "x20": RegisterDef(name="x20", size=8),
        "s4": RegisterAliasDef(name="s4", parent="x20", size=8, offset=0),
        # x21 is a callee-saved register
        "x21": RegisterDef(name="x21", size=8),
        "s5": RegisterAliasDef(name="s5", parent="x21", size=8, offset=0),
        # x22 is a callee-saved register
        "x22": RegisterDef(name="x22", size=8),
        "s6": RegisterAliasDef(name="s6", parent="x22", size=8, offset=0),
        # x23 is a callee-saved register
        "x23": RegisterDef(name="x23", size=8),
        "s7": RegisterAliasDef(name="s7", parent="x23", size=8, offset=0),
        # x24 is a callee-saved register
        "x24": RegisterDef(name="x24", size=8),
        "s8": RegisterAliasDef(name="s8", parent="x24", size=8, offset=0),
        # x25 is a callee-saved register
        "x25": RegisterDef(name="x25", size=8),
        "s9": RegisterAliasDef(name="s9", parent="x25", size=8, offset=0),
        # x26 is a callee-saved register
        "x26": RegisterDef(name="x26", size=8),
        "s10": RegisterAliasDef(name="s10", parent="x26", size=8, offset=0),
        # x27 is a callee-saved register
        "x27": RegisterDef(name="x27", size=8),
        "s11": RegisterAliasDef(name="s11", parent="x27", size=8, offset=0),
        # x28 is a temporary register
        "x28": RegisterDef(name="x28", size=8),
        "t3": RegisterAliasDef(name="t3", parent="x28", size=8, offset=0),
        # x29 is a temporary register
        "x29": RegisterDef(name="x29", size=8),
        "t4": RegisterAliasDef(name="t4", parent="x29", size=8, offset=0),
        # x30 is a temporary register
        "x30": RegisterDef(name="x30", size=8),
        "t5": RegisterAliasDef(name="t5", parent="x30", size=8, offset=0),
        # x31 is a temporary register
        "x31": RegisterDef(name="x31", size=8),
        "t6": RegisterAliasDef(name="t6", parent="x31", size=8, offset=0),
        # *** Program Counter ***
        "pc": RegisterDef(name="pc", size=8),
        # *** Floating-Point Registers ***
        # f0 is a temporary register
        "f0": RegisterDef(name="f0", size=8),
        "ft0": RegisterAliasDef(name="ft0", parent="f0", size=8, offset=0),
        # f1 is a temporary register
        "f1": RegisterDef(name="f1", size=8),
        "ft1": RegisterAliasDef(name="ft1", parent="f1", size=8, offset=0),
        # f2 is a temporary register
        "f2": RegisterDef(name="f2", size=8),
        "ft2": RegisterAliasDef(name="ft2", parent="f2", size=8, offset=0),
        # f3 is a temporary register
        "f3": RegisterDef(name="f3", size=8),
        "ft3": RegisterAliasDef(name="ft3", parent="f3", size=8, offset=0),
        # f4 is a temporary register
        "f4": RegisterDef(name="f4", size=8),
        "ft4": RegisterAliasDef(name="ft4", parent="f4", size=8, offset=0),
        # f5 is a temporary register
        "f5": RegisterDef(name="f5", size=8),
        "ft5": RegisterAliasDef(name="ft5", parent="f5", size=8, offset=0),
        # f6 is a temporary register
        "f6": RegisterDef(name="f6", size=8),
        "ft6": RegisterAliasDef(name="ft6", parent="f6", size=8, offset=0),
        # f7 is a temporary register
        "f7": RegisterDef(name="f7", size=8),
        "ft7": RegisterAliasDef(name="ft7", parent="f7", size=8, offset=0),
        # f8 is a callee saved register
        "f8": RegisterDef(name="f8", size=8),
        "fs0": RegisterAliasDef(name="fs0", parent="f8", size=8, offset=0),
        # f9 is a callee saved register
        "f9": RegisterDef(name="f9", size=8),
        "fs1": RegisterAliasDef(name="fs1", parent="f9", size=8, offset=0),
        # f10 is argument 0
        "f10": RegisterDef(name="f10", size=8),
        "fa0": RegisterAliasDef(name="fa0", parent="f10", size=8, offset=0),
        # f11 is argument 1
        "f11": RegisterDef(name="f11", size=8),
        "fa1": RegisterAliasDef(name="fa1", parent="f11", size=8, offset=0),
        # f12 is argument 2
        "f12": RegisterDef(name="f12", size=8),
        "fa2": RegisterAliasDef(name="fa2", parent="f12", size=8, offset=0),
        # f13 is argument 3
        "f13": RegisterDef(name="f13", size=8),
        "fa3": RegisterAliasDef(name="fa3", parent="f13", size=8, offset=0),
        # f14 is argument 4
        "f14": RegisterDef(name="f14", size=8),
        "fa4": RegisterAliasDef(name="fa4", parent="f14", size=8, offset=0),
        # f15 is argument 5
        "f15": RegisterDef(name="f15", size=8),
        "fa5": RegisterAliasDef(name="fa5", parent="f15", size=8, offset=0),
        # f16 is argument 6
        "f16": RegisterDef(name="f16", size=8),
        "fa6": RegisterAliasDef(name="fa6", parent="f16", size=8, offset=0),
        # f7 is argument 7
        "f17": RegisterDef(name="f17", size=8),
        "fa7": RegisterAliasDef(name="fa7", parent="f17", size=8, offset=0),
        # f18 is a callee-saved register
        "f18": RegisterDef(name="f18", size=8),
        "fs2": RegisterAliasDef(name="fs2", parent="f18", size=8, offset=0),
        # f19 is a callee-saved register
        "f19": RegisterDef(name="f19", size=8),
        "fs3": RegisterAliasDef(name="fs3", parent="f19", size=8, offset=0),
        # f20 is a callee-saved register
        "f20": RegisterDef(name="f20", size=8),
        "fs4": RegisterAliasDef(name="fs4", parent="f20", size=8, offset=0),
        # f21 is a callee-saved register
        "f21": RegisterDef(name="f21", size=8),
        "fs5": RegisterAliasDef(name="fs5", parent="f21", size=8, offset=0),
        # f22 is a callee-saved register
        "f22": RegisterDef(name="f22", size=8),
        "fs6": RegisterAliasDef(name="fs6", parent="f22", size=8, offset=0),
        # f23 is a callee-saved register
        "f23": RegisterDef(name="f23", size=8),
        "fs7": RegisterAliasDef(name="fs7", parent="f23", size=8, offset=0),
        # f24 is a callee-saved register
        "f24": RegisterDef(name="f24", size=8),
        "fs8": RegisterAliasDef(name="fs8", parent="f24", size=8, offset=0),
        # f25 is a callee-saved register
        "f25": RegisterDef(name="f25", size=8),
        "fs9": RegisterAliasDef(name="fs9", parent="f25", size=8, offset=0),
        # f26 is a callee-saved register
        "f26": RegisterDef(name="f26", size=8),
        "fs10": RegisterAliasDef(name="fs10", parent="f26", size=8, offset=0),
        # f27 is a callee-saved register
        "f27": RegisterDef(name="f27", size=8),
        "fs11": RegisterAliasDef(name="fs11", parent="f27", size=8, offset=0),
        # f28 is a temporary register
        "f28": RegisterDef(name="f28", size=8),
        "ft8": RegisterAliasDef(name="ft8", parent="f28", size=8, offset=0),
        # f29 is a temporary register
        "f29": RegisterDef(name="f29", size=8),
        "ft9": RegisterAliasDef(name="ft9", parent="f29", size=8, offset=0),
        # f30 is a temporary register
        "f30": RegisterDef(name="f30", size=8),
        "ft10": RegisterAliasDef(name="ft10", parent="f30", size=8, offset=0),
        # f31 is a temporary register
        "f31": RegisterDef(name="f31", size=8),
        "ft11": RegisterAliasDef(name="ft11", parent="f31", size=8, offset=0),
        # *** Vector Registers ***
        # NOTE: These exist, but are not supported
        # *** Control and Status Registers ***
        # NOTE: These exist, but aren't supported.
    }



import logging
import typing

from .. import hinting


class HintSubclassFilter(logging.Filter):
    """A custom logging filter based on Hint class."""

    def __init__(self, hint: typing.Type[hinting.Hint], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.hint = hint

    def filter(self, record):
        return isinstance(record.msg, self.hint)


__all__ = ["HintSubclassFilter"]



from ....platforms import Architecture, Byteorder
from .machdef import GhidraMachineDef


class i386MachineDef(GhidraMachineDef):
    arch = Architecture.X86_32
    byteorder = Byteorder.LITTLE
    language_id = "x86:LE:32:default"

    _registers = {
        # *** General Purpose Registers ***
        "eax": "eax",
        "ax": "ax",
        "al": "al",
        "ah": "ah",
        "ebx": "ebx",
        "bx": "bx",
        "bl": "bl",
        "bh": "bh",
        "ecx": "ecx",
        "cx": "cx",
        "cl": "cl",
        "ch": "ch",
        "edx": "edx",
        "dx": "dx",
        "dl": "dl",
        "dh": "dh",
        "esi": "esi",
        "si": "si",
        "sil": None,
        "edi": "edi",
        "di": "di",
        "dil": None,
        "ebp": "ebp",
        "bp": "bp",
        "bpl": None,
        "esp": "esp",
        "sp": "sp",
        "spl": None,
        # *** Instruction Pointer ***
        "eip": "eip",
        "ip": "ip",
        # *** Segment Registers ***
        "cs": "cs",
        "ds": "ds",
        "es": "es",
        "fs": "fs",
        "gs": "gs",
        "ss": "ss",
        # *** Flags Register ***
        "eflags": "eflags",
        "flags": "flags",
        # *** Control Registers ***
        "cr0": "cr0",
        "cr1": None,
        "cr2": "cr2",
        "cr3": "cr3",
        "cr4": "cr4",
        "cr8": None,
        # *** Debug Registers ***
        "dr0": "dr0",
        "dr1": "dr1",
        "dr2": "dr2",
        "dr3": "dr3",
        "dr6": "dr6",
        "dr7": "dr7",
        # *** Descriptor Table Registers ***
        "gdtr": "gdtr",
        "idtr": "idtr",
        "ldtr": "ldtr",
        # *** Task Register ***
        "tr": "tr",
        # *** x87 Registers ***
        # TODO: Ghidra seems to support x87, but I have no idea how its register file works
        # I can't find most of the control registers,
        # and there don't seem to be separate "fprN" registers; the stack references.
        "fpr0": None,
        "fpr1": None,
        "fpr2": None,
        "fpr3": None,
        "fpr4": None,
        "fpr5": None,
        "fpr6": None,
        "fpr7": None,
        "fctrl": None,
        "fstat": None,
        "ftag": None,
        "fip": None,
        "fdp": None,
        "fop": None,
        # *** MMX Registers ***
        "mm0": "mm0",
        "mm1": "mm1",
        "mm2": "mm2",
        "mm3": "mm3",
        "mm4": "mm4",
        "mm5": "mm5",
        "mm6": "mm6",
        "mm7": "mm7",
        # *** SSE Registers ***
        "xmm0": "xmm0",
        "xmm1": "xmm1",
        "xmm2": "xmm2",
        "xmm3": "xmm3",
        "xmm4": "xmm4",
        "xmm5": "xmm5",
        "xmm6": "xmm6",
        "xmm7": "xmm7",
    }



from .signal import *  # noqa F401,F403
from .signal import __all__ as __signal__
from .stdio import *  # noqa F401,F403
from .stdio import __all__ as __stdio__
from .stdlib import *  # noqa F401,F403
from .stdlib import __all__ as __stdlib__
from .string import *  # noqa F401,F403
from .string import __all__ as __string__
from .time import *  # noqa F401,F403
from .time import __all__ as __time__

__all__ = __signal__ + __stdlib__ + __string__ + __stdio__ + __time__
